# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, List, Optional, Union

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from mmengine.runner import set_random_seed
from PIL import Image
from tqdm.auto import tqdm

from mmagic.registry import MODELS
from mmagic.structures import DataSample
from mmagic.utils.typing import SampleList
from ..stable_diffusion.stable_diffusion import StableDiffusion
from .vico_utils import set_vico_modules

ModelType = Union[Dict, nn.Module]


@MODELS.register_module()
class ViCo(StableDiffusion):
    """Implementation of `ViCo with Stable Diffusion.

    <https://arxiv.org/abs/2306.00971>`_ (ViCo).

    Args:
        vae (Union[dict, nn.Module]): The config or module for VAE model.
        text_encoder (Union[dict, nn.Module]): The config or module for text
            encoder.
        tokenizer (str): The **name** for CLIP tokenizer.
        unet (Union[dict, nn.Module]): The config or module for Unet model.
        schedule (Union[dict, nn.Module]): The config or module for diffusion
            scheduler.
        test_scheduler (Union[dict, nn.Module], optional): The config or
            module for diffusion scheduler in test stage (`self.infer`). If not
            passed, will use the same scheduler as `schedule`. Defaults to
            None.
        val_prompts (Union[str, List[str]], optional): The prompts for
            validation. Defaults to None.
        num_class_images (int, optional): The number of images for class prior.
            Defaults to 3.
        dtype (str, optional): The dtype for the model. Defaults to 'fp16'.
        enable_xformers (bool, optional): Whether to use xformers.
            Defaults to True.
        noise_offset_weight (bool, optional): The weight of noise offset
            introduced in https://www.crosslabs.org/blog/diffusion-with-offset-noise  # noqa
            Defaults to 0.
        data_preprocessor (dict, optional): The pre-process config of
            :class:`BaseDataPreprocessor`. Defaults to
                dict(type='DataPreprocessor').
        init_cfg (dict, optional): The weight initialized config for
            :class:`BaseModule`. Defaults to None/
        image_cross_layers (List[int], optional): The layers to use image
            cross attention. Defaults to None.
        reg_loss_weight (float, optional): The weight of regularization loss.
            Defaults to 0.
        placeholder (str, optional): The placeholder token. Defaults to None.
        initialize_token (str, optional): The token to initialize the
            placeholder. Defaults to None.
        num_vectors_per_token (int, optional): The number of vectors per token.
    """

    def __init__(self,
                 vae: ModelType,
                 text_encoder: ModelType,
                 tokenizer: str,
                 unet: ModelType,
                 scheduler: ModelType,
                 test_scheduler: Optional[ModelType] = None,
                 val_prompts: Union[str, List[str]] = None,
                 dtype: str = 'fp16',
                 enable_xformers: bool = True,
                 noise_offset_weight: float = 0,
                 tomesd_cfg: Optional[dict] = None,
                 data_preprocessor: Optional[ModelType] = dict(
                     type='DataPreprocessor'),
                 init_cfg: Optional[dict] = None,
                 image_cross_layers: List[int] = None,
                 reg_loss_weight: float = 0,
                 placeholder: str = None,
                 initialize_token: str = None,
                 num_vectors_per_token: int = 1):

        super().__init__(vae, text_encoder, tokenizer, unet, scheduler,
                         test_scheduler, dtype, enable_xformers,
                         noise_offset_weight, tomesd_cfg, data_preprocessor,
                         init_cfg)
        self.reg_loss_weight = reg_loss_weight
        self.placeholder = placeholder

        self.dtype = torch.float32
        if dtype == 'fp16':
            self.dtype = torch.float16
        elif dtype == 'bf16':
            self.dtype = torch.bfloat16
        else:
            assert dtype in [
                'fp32', None
            ], ('dtype must be one of \'fp32\', \'fp16\', \'bf16\' or None.')

        self.val_prompts = val_prompts
        self.add_tokens(placeholder, initialize_token, num_vectors_per_token)
        self.set_vico(image_cross_layers)
        self.prepare_models()

    def prepare_models(self):
        """Prepare model for training.

        Move model to target dtype and disable gradient for some models.
        """
        """Disable gradient for untrainable modules to save memory."""
        self.vae.requires_grad_(False)
        self.unet.requires_grad_(False)
        self.text_encoder.set_only_embedding_trainable()
        self.set_only_imca_trainable()

    def set_vico(self, have_image_cross_attention: List[int]):
        """Set ViCo for model."""
        set_vico_modules(self.unet, have_image_cross_attention)

    def set_only_imca_trainable(self):
        """Set only image cross attention trainable."""
        for _, layer in self.unet.named_modules():
            if layer.__class__.__name__ == 'ViCoTransformer2D':
                if hasattr(layer, 'image_cross_attention'):
                    layer.image_cross_attention.train()
                    for name, param in (
                            layer.image_cross_attention.named_parameters()):
                        param.requires_grad = True

    def add_tokens(self,
                   placeholder_token: str,
                   initialize_token: str = None,
                   num_vectors_per_token: int = 1):
        """Add token for training.

        # TODO: support add tokens as dict, then we can load pretrained tokens.
        """
        self.tokenizer.add_placeholder_token(
            placeholder_token, num_vec_per_token=num_vectors_per_token)

        self.text_encoder.set_embedding_layer()
        embedding_layer = self.text_encoder.get_embedding_layer()
        assert embedding_layer is not None, (
            'Do not support get embedding layer for current text encoder. '
            'Please check your configuration.')

        if initialize_token:
            init_id = self.tokenizer(initialize_token).input_ids[1]
            initialize_embedding = embedding_layer.weight[init_id]
            initialize_embedding = initialize_embedding[None, ...].repeat(
                num_vectors_per_token, 1)
        else:
            emb_dim = embedding_layer.weight.shape[1]
            initialize_embedding = torch.zeros(num_vectors_per_token, emb_dim)

        token_info = self.tokenizer.get_token_info(placeholder_token)
        token_info['embedding'] = initialize_embedding
        token_info['trainable'] = True
        self.token_info = token_info
        embedding_layer.add_embeddings(token_info)

    @torch.no_grad()
    def val_step(self, data: dict) -> SampleList:
        """Gets the generated image of given data. Calls
        ``self.data_preprocessor`` and ``self.infer`` in order. Return the
        generated results which will be passed to evaluator or visualizer.

        Args:
            data (dict or tuple or list): Data sampled from dataset.

        Returns:
            SampleList: Generated image or image dict.
        """
        data = self.data_preprocessor(data)
        image_reference = data['inputs']['img_ref']
        data_samples = data['data_samples']
        if self.val_prompts is None:
            prompt = data_samples.prompt
        else:
            prompt = self.val_prompts
            # construct a fake data_sample for destruct
            data_samples.split() * len(prompt)
            data_samples = DataSample.stack(data_samples.split() * len(prompt))

        output = self.infer(prompt, image_reference, return_type='tensor')
        samples = output['samples']
        samples = self.data_preprocessor.destruct(samples, data_samples)

        out_data_sample = DataSample(fake_img=samples, prompt=prompt)
        data_sample_list = out_data_sample.split()
        return data_sample_list

    @torch.no_grad()
    def test_step(self, data: dict) -> SampleList:
        """Gets the generated image of given data. Calls
        ``self.data_preprocessor`` and ``self.infer`` in order. Return the
        generated results which will be passed to evaluator or visualizer.

        Args:
            data (dict or tuple or list): Data sampled from dataset.

        Returns:
            SampleList: Generated image or image dict.
        """
        data = self.data_preprocessor(data)
        image_reference = data['inputs']['img_ref']
        data_samples = data['data_samples']

        if self.val_prompts is None:
            prompt = data_samples.prompt
        else:
            prompt = self.val_prompts
            # construct a fake data_sample for destruct
            data_samples.split() * len(prompt)
            data_samples = DataSample.stack(data_samples.split() * len(prompt))

        output = self.infer(prompt, image_reference, return_type='tensor')
        samples = output['samples']
        samples = self.data_preprocessor.destruct(samples, data_samples)

        out_data_sample = DataSample(fake_img=samples, prompt=prompt)
        data_sample_list = out_data_sample.split()
        return data_sample_list

    def prepare_reference(self,
                          image_ref: Union[Image.Image, torch.Tensor],
                          height: Optional[int] = 512,
                          width: Optional[int] = 512):
        if isinstance(image_ref, Image.Image):
            if not image_ref.mode == 'RGB':
                image_ref = image_ref.convert('RGB')
            img = np.array(image_ref).astype(np.uint8)
            image_ref = Image.fromarray(img)
            image_ref = image_ref.resize((height, width),
                                         resample=Image.BILINEAR)

            image_ref = np.array(image_ref).astype(np.uint8)
            image_ref = (image_ref / 127.5 - 1.0).astype(np.float32)
            image_ref = torch.from_numpy(image_ref).permute(2, 0,
                                                            1).unsqueeze(0)

        return image_ref

    def train_step(self, data, optim_wrapper):
        """Training step."""
        data = self.data_preprocessor(data)
        inputs, data_samples = data['inputs'], data['data_samples']

        with optim_wrapper.optim_context(self.unet):
            image = inputs['img']  # image for new concept
            num_batches = image.shape[0]

            image_ref = inputs['img_ref']
            # cat image and image reference to avoid forward twice
            image = torch.cat([image, image_ref], dim=0)
            prompt_init = data_samples.prompt
            placeholder_string = self.placeholder

            image = image.to(self.dtype)
            latents = self.vae.encode(image).latent_dist.sample()
            latents = latents * self.vae.config.scaling_factor

            noise = torch.randn_like(latents[:num_batches, ...])
            timesteps = torch.randint(
                0,
                self.scheduler.num_train_timesteps, (num_batches, ),
                device=self.device)
            timesteps = timesteps.long()
            # image reference shares the same timesteps

            # only add noise to source image
            noisy_latents = self.scheduler.add_noise(
                latents[:num_batches, ...], noise, timesteps)
            noisy_latents = torch.cat(
                [noisy_latents, latents[num_batches:, ...]], dim=0)
            timesteps = torch.cat([timesteps, timesteps])

            input_ids = self.tokenizer(
                prompt_init,
                max_length=self.tokenizer.model_max_length,
                return_tensors='pt',
                padding='max_length',
                truncation=True)['input_ids'].to(self.device)
            ph_tokens = self.tokenizer(
                [placeholder_string],
                max_length=self.tokenizer.model_max_length,
                return_tensors='pt',
                padding='max_length',
                truncation=True)['input_ids'].to(self.device)
            ph_tok = ph_tokens[0, 1]
            placeholder_idx = torch.where(input_ids == ph_tok)
            clip_eot_token_id = self.tokenizer.encode(
                self.tokenizer.eos_token)['input_ids'][1]
            endoftext_idx = (torch.arange(input_ids.shape[0]),
                             (input_ids == clip_eot_token_id).nonzero(
                                 as_tuple=False)[0, 1])
            placeholder_position = [placeholder_idx, endoftext_idx]

            encoder_hidden_states = self.text_encoder(input_ids)[0]
            encoder_hidden_states = torch.cat(
                [encoder_hidden_states, encoder_hidden_states], dim=0)
            if self.scheduler.config.prediction_type == 'epsilon':
                gt = noise
            elif self.scheduler.config.prediction_type == 'v_prediction':
                gt = self.scheduler.get_velocity(latents, noise, timesteps)
            else:
                raise ValueError('Unknown prediction type '
                                 f'{self.scheduler.config.prediction_type}')

            # NOTE: we train unet in fp32, convert to float manually
            model_output = self.unet(
                noisy_latents.float(),
                timesteps,
                encoder_hidden_states=encoder_hidden_states.float(),
                placeholder_position=placeholder_position)
            model_pred = model_output['sample']
            loss_reg = model_output['loss_reg']

            loss_dict = dict()
            if loss_reg != 0:
                loss_dict['loss_reg'] = loss_reg * self.reg_loss_weight
            vico_loss = F.mse_loss(model_pred[:1].float(), gt.float())
            loss_dict['vico_loss'] = vico_loss
            parsed_loss, log_vars = self.parse_losses(loss_dict)
            optim_wrapper.update_params(parsed_loss)

        return log_vars

    @torch.no_grad()
    def infer(self,
              prompt: Union[str, List[str]],
              image_reference: Image.Image = None,
              height: Optional[int] = None,
              width: Optional[int] = None,
              num_inference_steps: int = 50,
              guidance_scale: float = 7.5,
              negative_prompt: Optional[Union[str, List[str]]] = None,
              num_images_per_prompt: Optional[int] = 1,
              eta: float = 0.0,
              generator: Optional[torch.Generator] = None,
              latents: Optional[torch.FloatTensor] = None,
              show_progress=True,
              seed=1,
              return_type='image'):
        """Function invoked when calling the pipeline for generation.

        Args:
            prompt (`str` or `List[str]`):
                The prompt or prompts to guide the image generation.
            height (`int`, *optional*,
                defaults to self.unet_sample_size * self.vae_scale_factor):
                The height in pixels of the generated image.
            width (`int`, *optional*,
                defaults to self.unet_sample_size * self.vae_scale_factor):
                The width in pixels of the generated image.
            num_inference_steps (`int`, *optional*, defaults to 50):
                The number of denoising steps.
                More denoising steps usually lead to a higher
                quality image at the expense of slower inference.
            guidance_scale (`float`, *optional*, defaults to 7.5):
                Guidance scale as defined in
                [Classifier-Free Diffusion Guidance]
                (https://arxiv.org/abs/2207.12598).
            negative_prompt (`str` or `List[str]`, *optional*):
                The prompt or prompts not to guide the image generation.
                Ignored when not using guidance (i.e., ignored
                if `guidance_scale` is less than `1`).
            num_images_per_prompt (`int`, *optional*, defaults to 1):
                The number of images to generate per prompt.
            eta (`float`, *optional*, defaults to 0.0):
                Corresponds to parameter eta (η) in the DDIM paper:
                https://arxiv.org/abs/2010.02502. Only applies to
                [`schedulers.DDIMScheduler`], will be ignored for others.
            generator (`torch.Generator`, *optional*):
                A [torch generator] to make generation deterministic.
            latents (`torch.FloatTensor`, *optional*):
                Pre-generated noisy latents,
                sampled from a Gaussian distribution,
                to be used as inputs for image generation.
                Can be used to tweak the same generation
                with different prompts.
                If not provided, a latents tensor will be
                generated by sampling using the supplied random `generator`.
            return_type (str): The return type of the inference results.
                Supported types are 'image', 'numpy', 'tensor'. If 'image'
                is passed, a list of PIL images will be returned. If 'numpy'
                is passed, a numpy array with shape [N, C, H, W] will be
                returned, and the value range will be same as decoder's
                output range. If 'tensor' is passed, the decoder's output
                will be returned. Defaults to 'image'.

        Returns:
            dict: A dict containing the generated images.
        """
        assert return_type in ['image', 'tensor', 'numpy']
        set_random_seed(seed=seed)

        # 0. Default height and width to unet
        height = height or self.unet_sample_size * self.vae_scale_factor
        width = width or self.unet_sample_size * self.vae_scale_factor

        # 1. Check inputs. Raise error if not correct
        self.check_inputs(prompt, height, width)

        # 2. Define call parameters
        batch_size = 1 if isinstance(prompt, str) else len(prompt)
        device = self.device

        img_dtype = self.vae.module.dtype if hasattr(self.vae, 'module') \
            else self.vae.dtype
        latent_dtype = next(self.unet.parameters()).dtype
        # here `guidance_scale` is defined analog to the
        # guidance weight `w` of equation (2)
        # of the Imagen paper: https://arxiv.org/pdf/2205.11487.pdf .
        # `guidance_scale = 1`
        # corresponds to doing no classifier free guidance.
        do_classifier_free_guidance = guidance_scale > 1.0

        # 3. Encode input prompt
        text_embeddings = self._encode_prompt(prompt, device,
                                              num_images_per_prompt,
                                              do_classifier_free_guidance,
                                              negative_prompt)
        uncond_embeddings, text_embeddings = text_embeddings.chunk(2)
        uncond_embeddings = torch.cat([uncond_embeddings] * 2)
        text_embeddings = torch.cat([text_embeddings] * 2)
        ph_tokens = self.tokenizer(
            num_images_per_prompt * [self.placeholder],
            max_length=self.tokenizer.model_max_length,
            return_tensors='pt',
            padding='max_length',
            truncation=True)['input_ids']
        input_ids = self.tokenizer(
            num_images_per_prompt * prompt,
            max_length=self.tokenizer.model_max_length,
            return_tensors='pt',
            padding='max_length',
            truncation=True)['input_ids']
        ph_tok = ph_tokens[0, 1]
        # TODO fix hard code
        clip_eot_token_id = 49407
        endoftext_idx = (torch.arange(input_ids.shape[0]),
                         torch.nonzero(input_ids == clip_eot_token_id)
                         [:batch_size, 1].repeat(num_images_per_prompt))
        placeholder_idx = torch.where(input_ids == ph_tok)
        if self.placeholder in prompt[0]:
            ph_pos = [placeholder_idx, endoftext_idx]
        else:
            ph_pos = [endoftext_idx, endoftext_idx]

        # 4. Prepare timesteps
        self.test_scheduler.set_timesteps(num_inference_steps)
        timesteps = self.test_scheduler.timesteps

        # 5. Prepare latent variables
        if hasattr(self.unet, 'module'):
            num_channels_latents = self.unet.module.in_channels
        else:
            num_channels_latents = self.unet.in_channels
        latents = self.prepare_latents(
            batch_size * num_images_per_prompt,
            num_channels_latents,
            height,
            width,
            text_embeddings.dtype,
            device,
            generator,
            latents,
        )
        image_reference = self.prepare_reference(
            image_reference,
            height,
            width,
        )
        image_reference = self.vae.encode(
            image_reference.to(dtype=img_dtype,
                               device=device)).latent_dist.sample()
        image_reference = image_reference.expand(
            batch_size * num_images_per_prompt, -1, -1, -1)
        image_reference = image_reference * self.vae.config.scaling_factor

        # 6. Prepare extra step kwargs.
        # TODO: Logic should ideally just be moved out of the pipeline
        extra_step_kwargs = self.prepare_extra_step_kwargs(generator, eta)

        # 7. Denoising loop
        if show_progress:
            timesteps = tqdm(timesteps)
        for i, t in enumerate(timesteps):
            latents = torch.cat([latents, image_reference], dim=0)
            latent_model_input = self.test_scheduler.scale_model_input(
                latents, t)
            latent_model_input = latent_model_input.to(latent_dtype)
            text_embeddings = text_embeddings.to(latent_dtype)
            uncond_embeddings = uncond_embeddings.to(latent_dtype)
            # predict the noise residual

            noise_pred_uncond = self.unet(
                latent_model_input,
                t,
                encoder_hidden_states=uncond_embeddings,
                placeholder_position=ph_pos)['sample'][:batch_size *
                                                       num_images_per_prompt]
            noise_pred_text = self.unet(
                latent_model_input,
                t,
                encoder_hidden_states=text_embeddings,
                placeholder_position=ph_pos)['sample'][:batch_size *
                                                       num_images_per_prompt]

            # perform guidance
            if do_classifier_free_guidance:
                noise_pred = noise_pred_uncond + guidance_scale * (
                    noise_pred_text - noise_pred_uncond)

                # compute the previous noisy sample x_t -> x_t-1
                latents_to_denoise = latents[:batch_size *
                                             num_images_per_prompt, ...]
                latents = self.test_scheduler.step(
                    noise_pred, t, latents_to_denoise,
                    **extra_step_kwargs)['prev_sample']

        # 8. Post-processing
        image = self.decode_latents(latents.to(img_dtype))
        if return_type == 'image':
            image = self.output_to_pil(image)
        elif return_type == 'numpy':
            image = image.cpu().numpy()
        else:
            assert return_type == 'tensor', (
                'Only support \'image\', \'numpy\' and \'tensor\' for '
                f'return_type, but receive {return_type}')

        return {'samples': image}

    def forward(self,
                inputs: torch.Tensor,
                data_samples: Optional[list] = None,
                mode: str = 'tensor') -> Union[Dict[str, torch.Tensor], list]:
        """forward is not implemented now."""
        raise NotImplementedError(
            'Forward is not implemented now, please use infer.')



# Copyright (c) OpenMMLab. All rights reserved.
import gradio as gr

from mmagic.apis import MMagicInferencer

editor = MMagicInferencer(model_name='controlnet_animation')


def process(video, prompt, a_prompt, negative_prompt,
            controlnet_conditioning_scale, width, height):
    prompt = prompt + a_prompt
    save_path = editor.infer(
        video=video,
        prompt=prompt,
        negative_prompt=negative_prompt,
        width=width,
        height=height,
        controlnet_conditioning_scale=controlnet_conditioning_scale)

    return save_path


block = gr.Blocks().queue()
with block:
    with gr.Row():
        gr.Markdown('## Controlnet Animation')
    with gr.Row():
        with gr.Column():
            video_inp = gr.Video(
                label='Video Source',
                source='upload',
                type='filepath',
                elem_id='input-vid')
            prompt = gr.Textbox(label='Prompt')
            run_button = gr.Button(label='Run')
            with gr.Accordion('Advanced options', open=False):
                a_prompt = gr.Textbox(
                    label='Added Prompt',
                    value='best quality, extremely detailed')
                n_prompt = gr.Textbox(
                    label='Negative Prompt',
                    value='longbody, lowres, bad anatomy, bad hands, ' +
                    'missing fingers, extra digit, fewer digits, '
                    'cropped, worst quality, low quality')
                controlnet_conditioning_scale = gr.Slider(
                    label='Control Weight',
                    minimum=0.0,
                    maximum=2.0,
                    value=0.7,
                    step=0.01)
                width = gr.Slider(
                    label='Image Width',
                    minimum=256,
                    maximum=768,
                    value=512,
                    step=64)
                height = gr.Slider(
                    label='Image Width',
                    minimum=256,
                    maximum=768,
                    value=512,
                    step=64)

        with gr.Column():
            video_out = gr.Video(label='Video Result', elem_id='video-output')
    ips = [
        video_inp, prompt, a_prompt, n_prompt, controlnet_conditioning_scale,
        width, height
    ]
    run_button.click(fn=process, inputs=ips, outputs=video_out)

block.launch(server_name='0.0.0.0', share=True)



# Copyright (c) OpenMMLab. All rights reserved.
import torch
import torch.nn as nn
import torch.nn.functional as F
from mmengine.model import BaseModule

from mmagic.models.archs import pixel_unshuffle
from mmagic.models.utils import make_layer
from mmagic.registry import MODELS


@MODELS.register_module()
class CAINNet(BaseModule):
    """CAIN network structure.

    Paper: Channel Attention Is All You Need for Video Frame Interpolation.
    Ref repo: https://github.com/myungsub/CAIN

    Args:
        in_channels (int): Channel number of inputs. Default: 3.
        kernel_size (int): Kernel size of CAINNet. Default: 3.
        num_block_groups (int): Number of block groups. Default: 5.
        num_block_layers (int): Number of blocks in a group. Default: 12.
        depth (int): Down scale depth, scale = 2**depth. Default: 3.
        reduction (int): Channel reduction of CA. Default: 16.
        norm (str | None): Normalization layer. If it is None, no
            normalization is performed. Default: None.
        padding (int): Padding of CAINNet. Default: 7.
        act (function): activate function. Default: nn.LeakyReLU(0.2, True).
        init_cfg (dict, optional): Initialization config dict. Default: None.
    """

    def __init__(self,
                 in_channels=3,
                 kernel_size=3,
                 num_block_groups=5,
                 num_block_layers=12,
                 depth=3,
                 reduction=16,
                 norm=None,
                 padding=7,
                 act=nn.LeakyReLU(0.2, True),
                 init_cfg=None):
        super().__init__(init_cfg=init_cfg)

        mid_channels = in_channels * (4**depth)
        self.scale = 2**depth
        self.padding = padding

        self.conv_first = nn.Conv2d(mid_channels * 2, mid_channels,
                                    kernel_size, 1, 1)
        self.body = make_layer(
            ResidualGroup,
            num_block_groups,
            block_layer=ResidualChannelAttention,
            num_block_layers=num_block_layers,
            mid_channels=mid_channels,
            kernel_size=kernel_size,
            reduction=reduction,
            norm=norm,
            act=act)
        self.conv_last = nn.Conv2d(mid_channels, mid_channels, kernel_size, 1,
                                   1)

    def forward(self, imgs, padding_flag=False):
        """Forward function.

        Args:
            imgs (Tensor): Input tensor with shape (n, 2, c, h, w).
            padding_flag (bool): Padding or not. Default: False.

        Returns:
            Tensor: Forward results.
        """

        assert imgs.shape[1] == 2
        x1, x2 = imgs[:, 0], imgs[:, 1]

        mean1 = x1.mean((2, 3), keepdim=True)
        mean2 = x2.mean((2, 3), keepdim=True)
        x1 -= mean1
        x2 -= mean2

        if padding_flag:
            padding_function, depadding_function = get_padding_functions(
                x1, self.padding)
            x1 = padding_function(x1)
            x2 = padding_function(x2)

        x1 = pixel_unshuffle(x1, self.scale)
        x2 = pixel_unshuffle(x2, self.scale)

        x = torch.cat([x1, x2], dim=1)
        x = self.conv_first(x)
        res = self.body(x)
        res += x
        x = self.conv_last(res)
        x = F.pixel_shuffle(x, self.scale)

        if padding_flag:
            x = depadding_function(x)

        x += (mean1 + mean2) / 2
        return x


def get_padding_functions(x, padding=7):
    """Generate padding function for CAIN.

    This function produces two functions to pad and depad a tensor, given the
    number of pixels to be padded. When applying padding and depadding
    sequentially, the original tensor is obtained.

    The generated padding function will pad the given tensor to the 'padding'
    power of 2, i.e., pow(2, 'padding').

    tensor --padding_function--> padded tensor
    padded tensor --depadding_function--> original tensor

    Args:
        x (Tensor): Input tensor.
        padding (int): Padding size. Default: 7.

    Returns:
        padding_function (Function): Padding function.
        depadding_function (Function): Depadding function.
    """

    h, w = x.shape[-2:]
    padding_width, padding_height = 0, 0
    if w != ((w >> padding) << padding):
        padding_width = (((w >> padding) + 1) << padding) - w
    if h != ((h >> padding) << padding):
        padding_height = (((h >> padding) + 1) << padding) - h

    left, right = padding_width // 2, padding_width - padding_width // 2
    up, down = padding_height // 2, padding_height - padding_height // 2
    if down >= h or right >= w:
        function = nn.ReplicationPad2d
    else:
        function = nn.ReflectionPad2d

    padding_function = function(padding=[left, right, up, down])
    depadding_function = function(
        padding=[0 - left, 0 - right, 0 - up, 0 - down])

    return padding_function, depadding_function


class ConvNormWithReflectionPad(BaseModule):
    """Apply reflection padding, followed by a convolution, which can be
    followed by an optional normalization.

    Args:
        in_channels (int): Channel number of input features.
        out_channels (int): Channel number of output features.
        kernel_size (int): Kernel size of convolution layer.
        norm (str | None): Normalization layer. If it is None, no
            normalization is performed. Default: None.
    """

    def __init__(self, in_channels, out_channels, kernel_size, norm=None):
        super().__init__()

        self.reflection_pad = nn.ReflectionPad2d(kernel_size // 2)
        self.conv = nn.Conv2d(
            in_channels, out_channels, kernel_size=kernel_size, bias=True)

        if norm is None:
            self.norm = None
        elif norm.lower() == 'in':
            self.norm = nn.InstanceNorm2d(
                out_channels, track_running_stats=True)
        elif norm.lower() == 'bn':
            self.norm = nn.BatchNorm2d(out_channels)
        else:
            raise ValueError(f"Invalid value for 'norm': {norm}")

    def forward(self, x):
        """Forward function for ConvNormWithReflectionPad.

        Args:
            x (Tensor): Input tensor with shape (n, c, h, w).

        Returns:
            Tensor: Output tensor with shape (n, c, h, w).
        """

        out = self.reflection_pad(x)
        out = self.conv(out)
        if self.norm:
            out = self.norm(out)
        return out


class ChannelAttentionLayer(BaseModule):
    """Channel Attention (CA) Layer.

    Args:
        mid_channels (int): Channel number of the intermediate features.
        reduction (int): Channel reduction of CA. Default: 16.
    """

    def __init__(self, mid_channels, reduction=16):
        super().__init__()

        # global average pooling: (n, c, h, w) --> (n, c, 1, 1)
        self.avg_pool = nn.AdaptiveAvgPool2d(1)
        # channel reduction.
        self.channel_attention = nn.Sequential(
            nn.Conv2d(
                mid_channels,
                mid_channels // reduction,
                1,
                padding=0,
                bias=True), nn.ReLU(inplace=True),
            nn.Conv2d(
                mid_channels // reduction,
                mid_channels,
                1,
                padding=0,
                bias=True), nn.Sigmoid())

    def forward(self, x):
        """Forward function for ChannelAttentionLayer.

        Args:
            x (Tensor): Input tensor with shape (n, c, h, w).

        Returns:
            Tensor: Output tensor with shape (n, c, h, w).
        """

        y = self.avg_pool(x)
        y = self.channel_attention(y)
        return x * y


class ResidualChannelAttention(BaseModule):
    """Residual Channel Attention Module.

    Args:
        mid_channels (int): Channel number of the intermediate features.
        kernel_size (int): Kernel size of convolution layers. Default: 3.
        reduction (int): Channel reduction. Default: 16.
        norm (None | function): Norm layer. If None, no norm layer.
            Default: None.
        act (function): activation function. Default: nn.LeakyReLU(0.2, True).
    """

    def __init__(self,
                 mid_channels,
                 kernel_size=3,
                 reduction=16,
                 norm=None,
                 act=nn.LeakyReLU(0.2, True)):
        super().__init__()

        self.body = nn.Sequential(
            ConvNormWithReflectionPad(
                mid_channels, mid_channels, kernel_size, norm=norm), act,
            ConvNormWithReflectionPad(
                mid_channels, mid_channels, kernel_size, norm=norm),
            ChannelAttentionLayer(mid_channels, reduction))

    def forward(self, x):
        """Forward function for ResidualChannelAttention.

        Args:
            x (Tensor): Input tensor with shape (n, c, h, w).

        Returns:
            Tensor: Output tensor with shape (n, c, h, w).
        """

        out = self.body(x)
        return out + x


class ResidualGroup(BaseModule):
    """Residual Group, consisting of a stack of residual channel attention,
    followed by a convolution.

    Args:
        block_layer (nn.Module): nn.Module class for basic block.
        num_block_layers (int): number of blocks.
        mid_channels (int): Channel number of the intermediate features.
        kernel_size (int): Kernel size of ResidualGroup.
        reduction (int): Channel reduction of CA. Default: 16.
        act (function): activation function. Default: nn.LeakyReLU(0.2, True).
        norm (str | None): Normalization layer. If it is None, no
            normalization is performed. Default: None.
    """

    def __init__(self,
                 block_layer,
                 num_block_layers,
                 mid_channels,
                 kernel_size,
                 reduction,
                 act=nn.LeakyReLU(0.2, True),
                 norm=None):
        super().__init__()

        self.body = make_layer(
            block_layer,
            num_block_layers,
            mid_channels=mid_channels,
            kernel_size=kernel_size,
            reduction=reduction,
            norm=norm,
            act=act)
        self.conv_after_body = ConvNormWithReflectionPad(
            mid_channels, mid_channels, kernel_size, norm=norm)

    def forward(self, x):
        """Forward function for ResidualGroup.

        Args:
            x (Tensor): Input tensor with shape (n, c, h, w).

        Returns:
            Tensor: Output tensor with shape (n, c, h, w).
        """

        y = self.body(x)
        y = self.conv_after_body(y)
        return x + y



# Copyright (c) OpenMMLab. All rights reserved.
import torch
import torch.nn.functional as F
from mmengine.model import BaseModule

from mmagic.registry import MODELS


@MODELS.register_module()
class SearchTransformer(BaseModule):
    """Search texture reference by transformer.

    Include relevance embedding, hard-attention and soft-attention.
    """

    def gather(self, inputs, dim, index):
        """Hard Attention. Gathers values along an axis specified by dim.

        Args:
            inputs (Tensor): The source tensor. (N, C*k*k, H*W)
            dim (int): The axis along which to index.
            index (Tensor): The indices of elements to gather. (N, H*W)

        results:
            outputs (Tensor): The result tensor. (N, C*k*k, H*W)
        """

        views = [inputs.size(0)
                 ] + [1 if i != dim else -1 for i in range(1, inputs.ndim)]
        expansion = [
            -1 if i in (0, dim) else d for i, d in enumerate(inputs.size())
        ]
        index = index.view(views).expand(expansion)
        outputs = torch.gather(inputs, dim, index)

        return outputs

    def forward(self, img_lq, ref_lq, refs):
        """Texture transformer.

        Q = LTE(img_lq)
        K = LTE(ref_lq)
        V = LTE(ref), from V_level_n to V_level_1

        Relevance embedding aims to embed the relevance between the LQ and
            Ref image by estimating the similarity between Q and K.
        Hard-Attention: Only transfer features from the most relevant position
            in V for each query.
        Soft-Attention: synthesize features from the transferred GT texture
            features T and the LQ features F from the backbone.

        Args:
            All args are features come from extractor (such as LTE).
                These features contain 3 levels.
                When upscale_factor=4, the size ratio of these features is
                level3:level2:level1 = 1:2:4.
            img_lq (Tensor): Tensor of 4x bicubic-upsampled lq image.
                (N, C, H, W)
            ref_lq (Tensor): Tensor of ref_lq. ref_lq is obtained
                by applying bicubic down-sampling and up-sampling with factor
                4x on ref. (N, C, H, W)
            refs (Tuple[Tensor]): Tuple of ref tensors.
                [(N, C, H, W), (N, C/2, 2H, 2W), ...]

        Returns:
            tuple: tuple contains:
                soft_attention (Tensor): Soft-Attention tensor. (N, 1, H, W) \n
                textures (Tuple[Tensor]): Transferred GT textures.
                [(N, C, H, W), (N, C/2, 2H, 2W), ...]
        """

        levels = len(refs)
        # query
        query = F.unfold(img_lq, kernel_size=(3, 3), padding=1)

        # key
        key = F.unfold(ref_lq, kernel_size=(3, 3), padding=1)
        key_t = key.permute(0, 2, 1)

        # values
        values = [
            F.unfold(
                refs[i],
                kernel_size=3 * pow(2, i),
                padding=pow(2, i),
                stride=pow(2, i)) for i in range(levels)
        ]

        key_t = F.normalize(key_t, dim=2)  # [N, H*W, C*k*k]
        query = F.normalize(query, dim=1)  # [N, C*k*k, H*W]

        # Relevance embedding
        rel_embedding = torch.bmm(key_t, query)  # [N, H*W, H*W]
        max_val, max_index = torch.max(rel_embedding, dim=1)  # [N, H*W]

        # hard-attention
        textures = [self.gather(value, 2, max_index) for value in values]

        # to tensor
        h, w = img_lq.size()[-2:]
        textures = [
            F.fold(
                textures[i],
                output_size=(h * pow(2, i), w * pow(2, i)),
                kernel_size=3 * pow(2, i),
                padding=pow(2, i),
                stride=pow(2, i)) / 9. for i in range(levels)
        ]

        soft_attention = max_val.view(max_val.size(0), 1, h, w)

        return soft_attention, textures



# Copyright (c) OpenMMLab. All rights reserved.
from .aot_decoder import AOTDecoder
from .aot_encoder import AOTEncoder
from .aot_encoder_decoder import AOTEncoderDecoder
from .aot_inpaintor import AOTInpaintor
from .aot_neck import AOTBlockNeck

__all__ = [
    'AOTEncoderDecoder', 'AOTBlockNeck', 'AOTInpaintor', 'AOTEncoder',
    'AOTDecoder'
]



# Copyright (c) OpenMMLab. All rights reserved.
import torch
import torch.nn as nn
from mmcv.cnn import ConvModule
from mmengine.model.weight_init import constant_init, xavier_init
from torch.nn import functional as F


class GCAModule(nn.Module):
    """Guided Contextual Attention Module.

    From https://arxiv.org/pdf/2001.04069.pdf.
    Based on https://github.com/nbei/Deep-Flow-Guided-Video-Inpainting.
    This module use image feature map to augment the alpha feature map with
    guided contextual attention score.

    Image feature and alpha feature are unfolded to small patches and later
    used as conv kernel. Thus, we refer the unfolding size as kernel size.
    Image feature patches have a default kernel size 3 while the kernel size of
    alpha feature patches could be specified by `rate` (see `rate` below). The
    image feature patches are used to convolve with the image feature itself
    to calculate the contextual attention. Then the attention feature map is
    convolved by alpha feature patches to obtain the attention alpha feature.
    At last, the attention alpha feature is added to the input alpha feature.

    Args:
        in_channels (int): Input channels of the guided contextual attention
            module.
        out_channels (int): Output channels of the guided contextual attention
            module.
        kernel_size (int): Kernel size of image feature patches. Default 3.
        stride (int): Stride when unfolding the image feature. Default 1.
        rate (int): The downsample rate of image feature map. The corresponding
            kernel size and stride of alpha feature patches will be `rate x 2`
            and `rate`. It could be regarded as the granularity of the gca
            module. Default: 2.
        pad_args (dict): Parameters of padding when convolve image feature with
            image feature patches or alpha feature patches. Allowed keys are
            `mode` and `value`. See torch.nn.functional.pad() for more
            information. Default: dict(mode='reflect').
        interpolation (str): Interpolation method in upsampling and
            downsampling.
        penalty (float): Punishment hyperparameter to avoid a large correlation
            between each unknown patch and itself. Default: -1e4.
        eps (float): A small number to avoid dividing by 0 when calculating
            the normed image feature patch. Default: 1e-4.
    """

    def __init__(self,
                 in_channels,
                 out_channels,
                 kernel_size=3,
                 stride=1,
                 rate=2,
                 pad_args=dict(mode='reflect'),
                 interpolation='nearest',
                 penalty=-1e4,
                 eps=1e-4):
        super().__init__()
        self.kernel_size = kernel_size
        self.stride = stride
        self.rate = rate
        self.pad_args = pad_args
        self.interpolation = interpolation
        self.penalty = penalty
        self.eps = eps

        # reduced the channels of input image feature.
        self.guidance_conv = nn.Conv2d(in_channels, in_channels // 2, 1)

        # convolution after the attention alpha feature
        self.out_conv = ConvModule(
            out_channels,
            out_channels,
            1,
            norm_cfg=dict(type='BN'),
            act_cfg=None)

        self.init_weights()

    def init_weights(self):
        """Init weights for the model."""
        xavier_init(self.guidance_conv, distribution='uniform')
        xavier_init(self.out_conv.conv, distribution='uniform')
        constant_init(self.out_conv.norm, 1e-3)

    def forward(self, img_feat, alpha_feat, unknown=None, softmax_scale=1.):
        """Forward function of GCAModule.

        Args:
            img_feat (Tensor): Image feature map of shape
                (N, ori_c, ori_h, ori_w).
            alpha_feat (Tensor): Alpha feature map of shape
                (N, alpha_c, ori_h, ori_w).
            unknown (Tensor, optional): Unknown area map generated by trimap.
                If specified, this tensor should have shape
                (N, 1, ori_h, ori_w).
            softmax_scale (float, optional): The softmax scale of the attention
                if unknown area is not provided in forward. Default: 1.

        Returns:
            Tensor: The augmented alpha feature.
        """

        if alpha_feat.shape[2:4] != img_feat.shape[2:4]:
            raise ValueError(
                'image feature size does not align with alpha feature size: '
                f'image feature size {img_feat.shape[2:4]}, '
                f'alpha feature size {alpha_feat.shape[2:4]}')

        if unknown is not None and unknown.shape[2:4] != img_feat.shape[2:4]:
            raise ValueError(
                'image feature size does not align with unknown mask size: '
                f'image feature size {img_feat.shape[2:4]}, '
                f'unknown mask size {unknown.shape[2:4]}')

        # preprocess image feature
        img_feat = self.guidance_conv(img_feat)
        img_feat = F.interpolate(
            img_feat, scale_factor=1 / self.rate, mode=self.interpolation)

        # preprocess unknown mask
        unknown, softmax_scale = self.process_unknown_mask(
            unknown, img_feat, softmax_scale)

        img_ps, alpha_ps, unknown_ps = self.extract_feature_maps_patches(
            img_feat, alpha_feat, unknown)

        # create self correlation mask with shape:
        # (N, img_h*img_w, img_h, img_w)
        self_mask = self.get_self_correlation_mask(img_feat)

        # split tensors by batch dimension; tuple is returned
        img_groups = torch.split(img_feat, 1, dim=0)
        img_ps_groups = torch.split(img_ps, 1, dim=0)
        alpha_ps_groups = torch.split(alpha_ps, 1, dim=0)
        unknown_ps_groups = torch.split(unknown_ps, 1, dim=0)
        scale_groups = torch.split(softmax_scale, 1, dim=0)
        groups = (img_groups, img_ps_groups, alpha_ps_groups,
                  unknown_ps_groups, scale_groups)

        out = []
        # i is the virtual index of the sample in the current batch
        for img_i, img_ps_i, alpha_ps_i, unknown_ps_i, scale_i in zip(*groups):
            similarity_map = self.compute_similarity_map(img_i, img_ps_i)

            gca_score = self.compute_guided_attention_score(
                similarity_map, unknown_ps_i, scale_i, self_mask)

            out_i = self.propagate_alpha_feature(gca_score, alpha_ps_i)

            out.append(out_i)

        out = torch.cat(out, dim=0)
        out.reshape_as(alpha_feat)

        out = self.out_conv(out) + alpha_feat
        return out

    def extract_feature_maps_patches(self, img_feat, alpha_feat, unknown):
        """Extract image feature, alpha feature unknown patches.

        Args:
            img_feat (Tensor): Image feature map of shape
                (N, img_c, img_h, img_w).
            alpha_feat (Tensor): Alpha feature map of shape
                (N, alpha_c, ori_h, ori_w).
            unknown (Tensor, optional): Unknown area map generated by trimap of
                shape (N, 1, img_h, img_w).

        Returns:
            tuple: 3-tuple of

                ``Tensor``: Image feature patches of shape \
                    (N, img_h*img_w, img_c, img_ks, img_ks).

                ``Tensor``: Guided contextual attention alpha feature map. \
                    (N, img_h*img_w, alpha_c, alpha_ks, alpha_ks).

                ``Tensor``: Unknown mask of shape (N, img_h*img_w, 1, 1).
        """
        # extract image feature patches with shape:
        # (N, img_h*img_w, img_c, img_ks, img_ks)
        img_ks = self.kernel_size
        img_ps = self.extract_patches(img_feat, img_ks, self.stride)

        # extract alpha feature patches with shape:
        # (N, img_h*img_w, alpha_c, alpha_ks, alpha_ks)
        alpha_ps = self.extract_patches(alpha_feat, self.rate * 2, self.rate)

        # extract unknown mask patches with shape: (N, img_h*img_w, 1, 1)
        unknown_ps = self.extract_patches(unknown, img_ks, self.stride)
        unknown_ps = unknown_ps.squeeze(dim=2)  # squeeze channel dimension
        unknown_ps = unknown_ps.mean(dim=[2, 3], keepdim=True)

        return img_ps, alpha_ps, unknown_ps

    def compute_similarity_map(self, img_feat, img_ps):
        """Compute similarity between image feature patches.

        Args:
            img_feat (Tensor): Image feature map of shape
                (1, img_c, img_h, img_w).
            img_ps (Tensor): Image feature patches tensor of shape
                (1, img_h*img_w, img_c, img_ks, img_ks).

        Returns:
            Tensor: Similarity map between image feature patches with shape \
                (1, img_h*img_w, img_h, img_w).
        """
        img_ps = img_ps[0]  # squeeze dim 0
        # convolve the feature to get correlation (similarity) map
        escape_NaN = torch.FloatTensor([self.eps]).to(img_feat)
        img_ps_normed = img_ps / torch.max(self.l2_norm(img_ps), escape_NaN)
        img_feat = self.pad(img_feat, self.kernel_size, self.stride)
        similarity_map = F.conv2d(img_feat, img_ps_normed)

        return similarity_map

    def compute_guided_attention_score(self, similarity_map, unknown_ps, scale,
                                       self_mask):
        """Compute guided attention score.

        Args:
            similarity_map (Tensor): Similarity map of image feature with shape
                (1, img_h*img_w, img_h, img_w).
            unknown_ps (Tensor): Unknown area patches tensor of shape
                (1, img_h*img_w, 1, 1).
            scale (Tensor): Softmax scale of known and unknown area:
                [unknown_scale, known_scale].
            self_mask (Tensor): Self correlation mask of shape
                (1, img_h*img_w, img_h, img_w). At (1, i*i, i, i) mask value
                equals -1e4 for i in [1, img_h*img_w] and other area is all
                zero.

        Returns:
            Tensor: Similarity map between image feature patches with shape \
                (1, img_h*img_w, img_h, img_w).
        """
        # scale the correlation with predicted scale factor for known and
        # unknown area
        unknown_scale, known_scale = scale[0]
        out = similarity_map * (
            unknown_scale * unknown_ps.gt(0.).float() +
            known_scale * unknown_ps.le(0.).float())
        # mask itself, self-mask only applied to unknown area
        out = out + self_mask * unknown_ps
        gca_score = F.softmax(out, dim=1)

        return gca_score

    def propagate_alpha_feature(self, gca_score, alpha_ps):
        """Propagate alpha feature based on guided attention score.

        Args:
            gca_score (Tensor): Guided attention score map of shape
                (1, img_h*img_w, img_h, img_w).
            alpha_ps (Tensor): Alpha feature patches tensor of shape
                (1, img_h*img_w, alpha_c, alpha_ks, alpha_ks).

        Returns:
            Tensor: Propagated alpha feature map of shape \
                (1, alpha_c, alpha_h, alpha_w).
        """
        alpha_ps = alpha_ps[0]  # squeeze dim 0
        if self.rate == 1:
            gca_score = self.pad(gca_score, kernel_size=2, stride=1)
            alpha_ps = alpha_ps.permute(1, 0, 2, 3)
            out = F.conv2d(gca_score, alpha_ps) / 4.
        else:
            out = F.conv_transpose2d(
                gca_score, alpha_ps, stride=self.rate, padding=1) / 4.

        return out

    def process_unknown_mask(self, unknown, img_feat, softmax_scale):
        """Process unknown mask.

        Args:
            unknown (Tensor, optional): Unknown area map generated by trimap of
                shape (N, 1, ori_h, ori_w)
            img_feat (Tensor): The interpolated image feature map of shape
                (N, img_c, img_h, img_w).
            softmax_scale (float, optional): The softmax scale of the attention
                if unknown area is not provided in forward. Default: 1.

        Returns:
            tuple: 2-tuple of

                ``Tensor``: Interpolated unknown area map of shape \
                    (N, img_h*img_w, img_h, img_w).

                ``Tensor``: Softmax scale tensor of known and unknown area of \
                    shape (N, 2).
        """
        n, _, h, w = img_feat.shape

        if unknown is not None:
            unknown = unknown.clone()
            unknown = F.interpolate(
                unknown, scale_factor=1 / self.rate, mode=self.interpolation)
            unknown_mean = unknown.mean(dim=[2, 3])
            known_mean = 1 - unknown_mean
            unknown_scale = torch.clamp(
                torch.sqrt(unknown_mean / known_mean), 0.1, 10).to(img_feat)
            known_scale = torch.clamp(
                torch.sqrt(known_mean / unknown_mean), 0.1, 10).to(img_feat)
            softmax_scale = torch.cat([unknown_scale, known_scale], dim=1)
        else:
            unknown = torch.ones((n, 1, h, w)).to(img_feat)
            softmax_scale = torch.FloatTensor(
                [softmax_scale,
                 softmax_scale]).view(1, 2).repeat(n, 1).to(img_feat)

        return unknown, softmax_scale

    def extract_patches(self, x, kernel_size, stride):
        """Extract feature patches.

        The feature map will be padded automatically to make sure the number of
        patches is equal to `(H / stride) * (W / stride)`.

        Args:
            x (Tensor): Feature map of shape (N, C, H, W).
            kernel_size (int): Size of each patches.
            stride (int): Stride between patches.

        Returns:
            Tensor: Extracted patches of shape \
                (N, (H / stride) * (W / stride) , C, kernel_size, kernel_size).
        """
        n, c, _, _ = x.shape
        x = self.pad(x, kernel_size, stride)
        x = F.unfold(x, (kernel_size, kernel_size), stride=(stride, stride))
        x = x.permute(0, 2, 1)
        x = x.reshape(n, -1, c, kernel_size, kernel_size)
        return x

    def pad(self, x, kernel_size, stride):
        """Pad input tensor.

        Args:
            x (Tensor): Input tensor.
            kernel_size (int): Kernel size of conv layer.
            stride (int): Stride of conv layer.

        Returns:
            Tensor: Padded tensor
        """
        left = (kernel_size - stride + 1) // 2
        right = (kernel_size - stride) // 2
        pad = (left, right, left, right)
        return F.pad(x, pad, **self.pad_args)

    def get_self_correlation_mask(self, img_feat):
        """Create self correlation mask.

        Args:
            img_feat (Tensor): Input tensor.

        Returns:
            Tensor: Mask tensor.
        """
        _, _, h, w = img_feat.shape
        # As ONNX does not support dynamic num_classes, we have to convert it
        # into an integer
        self_mask = F.one_hot(
            torch.arange(h * w).view(h, w), num_classes=int(h * w))
        self_mask = self_mask.permute(2, 0, 1).view(1, h * w, h, w)
        # use large negative value to mask out self-correlation before softmax
        self_mask = self_mask * self.penalty
        return self_mask.to(img_feat)

    @staticmethod
    def l2_norm(x):
        """L2 normalization function.

        Args:
            x (Tensor): Input tensor.

        Returns:
            Tensor: L2 normalized output tensor.
        """

        x = x**2
        x = x.sum(dim=[1, 2, 3], keepdim=True)
        return torch.sqrt(x)



# Copyright (c) OpenMMLab. All rights reserved.
import gc
import types
from collections import OrderedDict
from typing import Optional, Tuple, Union

import torch
import torch.nn.functional as F
import torchvision.transforms as T
from torch import nn
from torch.nn import Linear
from transformers import (CLIPModel, CLIPPreTrainedModel, CLIPTextModel,
                          CLIPVisionConfig, CLIPVisionModel)
from transformers.modeling_outputs import BaseModelOutputWithPooling
from transformers.models.clip.modeling_clip import _expand_mask


class FastComposerModel(nn.Module):
    """FastComposerModel is based on the StableDiffusion Model and the Clip
    Model."""

    def __init__(self, text_encoder, image_encoder, vae, unet, cfg):
        super().__init__()
        self.text_encoder = text_encoder
        self.image_encoder = image_encoder
        self.vae = vae
        self.unet = unet
        self.use_ema = False
        self.ema_param = None
        self.pretrained_model_name_or_path = cfg[
            'pretrained_model_name_or_path']
        self.revision = cfg['revision']
        self.non_ema_revision = cfg['non_ema_revision']
        self.object_localization = cfg['object_localization']
        self.object_localization_weight = cfg['object_localization_weight']
        self.localization_layers = cfg['localization_layers']
        self.mask_loss = cfg['mask_loss']
        self.mask_loss_prob = cfg['mask_loss_prob']

        embed_dim = text_encoder.config.hidden_size

        self.postfuse_module = FastComposerPostfuseModule(embed_dim)

        if self.object_localization:
            self.cross_attention_scores = {}
            self.unet = unet_store_cross_attention_scores(
                self.unet, self.cross_attention_scores,
                self.localization_layers)
            self.object_localization_loss_fn = BalancedL1Loss(
                cfg['object_localization_threshold'],
                cfg['object_localization_normalize'],
            )

    def _clear_cross_attention_scores(self):
        """Delete cross attention scores."""
        if hasattr(self, 'cross_attention_scores'):
            keys = list(self.cross_attention_scores.keys())
            for k in keys:
                del self.cross_attention_scores[k]

        gc.collect()

    @staticmethod
    def from_pretrained(cfg, vae, unet):
        """Init FastComposerTextEncoder and FastComposerCLIPImageEncoder."""

        text_encoder = FastComposerTextEncoder.from_pretrained(
            cfg['pretrained_model_name_or_path'],
            subfolder='text_encoder',
            revision=cfg['revision'],
        )
        if not isinstance(cfg['image_encoder'], dict):
            image_encoder = FastComposerCLIPImageEncoder.from_pretrained(
                cfg['image_encoder'])
        else:
            vision_model = CLIPVisionModel(
                CLIPVisionConfig.from_dict(cfg['image_encoder']))
            visual_projection = Linear(
                in_features=1024, out_features=768, bias=False)
            vision_processor = T.Normalize(
                (0.48145466, 0.4578275, 0.40821073),
                (0.26862954, 0.26130258, 0.27577711),
            )
            image_encoder = FastComposerCLIPImageEncoder(
                vision_model,
                visual_projection,
                vision_processor,
            )

        return FastComposerModel(text_encoder, image_encoder, vae, unet, cfg)

    def forward(self, batch, noise_scheduler):
        """Forward function.

        Args:
            batch (torch.Tensor ):
                You can directly input a ``torch.Tensor``.
            noise_scheduler (torch.Tensor ):
                You can directly input a ``torch.Tensor``.

        Returns:
            Dict
        """

        pixel_values = batch['pixel_values']
        input_ids = batch['input_ids']
        image_token_mask = batch['image_token_mask']
        object_pixel_values = batch['object_pixel_values']
        num_objects = batch['num_objects']

        vae_dtype = self.vae.parameters().__next__().dtype
        vae_input = pixel_values.to(vae_dtype)

        latents = self.vae.encode(vae_input).latent_dist.sample()
        latents = latents * self.vae.config.scaling_factor

        # Sample noise that we'll add to the latents
        noise = torch.randn_like(latents)
        bsz = latents.shape[0]
        # Sample a random timestep for each image
        timesteps = torch.randint(
            0,
            noise_scheduler.num_train_timesteps, (bsz, ),
            device=latents.device)
        timesteps = timesteps.long()

        # Add noise to the latents according to
        # the noise magnitude at each timestep
        # (this is the forward diffusion process)
        noisy_latents = noise_scheduler.add_noise(latents, noise, timesteps)

        # (bsz, max_num_objects, num_image_tokens, dim)
        object_embeds = self.image_encoder(object_pixel_values)

        encoder_hidden_states = self.text_encoder(
            input_ids, image_token_mask, object_embeds,
            num_objects)[0]  # (bsz, seq_len, dim)

        encoder_hidden_states = self.postfuse_module(
            encoder_hidden_states,
            object_embeds,
            image_token_mask,
            num_objects,
        )

        # Get the target for loss depending on the prediction type
        if noise_scheduler.config.prediction_type == 'epsilon':
            target = noise
        elif noise_scheduler.config.prediction_type == 'v_prediction':
            target = noise_scheduler.get_velocity(latents, noise, timesteps)
        else:
            raise ValueError('Unknown prediction type '
                             f'{noise_scheduler.config.prediction_type}')

        pred = self.unet(noisy_latents, timesteps,
                         encoder_hidden_states).sample

        if self.mask_loss and torch.rand(1) < self.mask_loss_prob:
            object_segmaps = batch['object_segmaps']
            mask = (object_segmaps.sum(dim=1) > 0).float()
            mask = F.interpolate(
                mask.unsqueeze(1),
                size=(pred.shape[-2], pred.shape[-1]),
                mode='bilinear',
                align_corners=False,
            )
            pred = pred * mask
            target = target * mask

        denoise_loss = F.mse_loss(
            pred.float(), target.float(), reduction='mean')

        return_dict = {'denoise_loss': denoise_loss}

        if self.object_localization:
            object_segmaps = batch['object_segmaps']
            image_token_idx = batch['image_token_idx']
            image_token_idx_mask = batch['image_token_idx_mask']
            localization_loss = get_object_localization_loss(
                self.cross_attention_scores,
                object_segmaps,
                image_token_idx,
                image_token_idx_mask,
                self.object_localization_loss_fn,
            )
            return_dict['localization_loss'] = localization_loss
            loss = self.object_localization_weight * localization_loss
            loss += denoise_loss
            self._clear_cross_attention_scores()
        else:
            loss = denoise_loss

        return_dict['loss'] = loss
        return return_dict


class FastComposerTextEncoder(CLIPPreTrainedModel):
    """TextEncoder for FastComposerModel."""

    @staticmethod
    def from_pretrained(model_name_or_path, **kwargs):
        """Init textEncoder with Stable Diffusion Model name or path."""
        model = CLIPTextModel.from_pretrained(model_name_or_path, **kwargs)
        text_model = model.text_model
        return FastComposerTextEncoder(text_model)

    def __init__(self, text_model):
        super().__init__(text_model.config)
        self.config = text_model.config
        self.final_layer_norm = text_model.final_layer_norm
        self.embeddings = text_model.embeddings
        self.encoder = text_model.encoder
        self._build_causal_attention_mask = build_causal_attention_mask

    def forward(
        self,
        input_ids,
        image_token_mask=None,
        object_embeds=None,
        num_objects=None,
        attention_mask: Optional[torch.Tensor] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ) -> Union[Tuple, BaseModelOutputWithPooling]:
        """Forward function.

        Args:
            input_ids (torch.Tensor ):
                You can directly input a ``torch.Tensor``.
            image_token_mask (torch.Tensor ):
                You can directly input a ``torch.Tensor``.
            object_embeds (torch.Tensor ):
                You can directly input a ``torch.Tensor``.
            num_objects (torch.Tensor ):
                You can directly input a ``torch.Tensor``.
            attention_mask (torch.Tensor ):
                You can directly input a ``torch.Tensor``.
            output_attentions (bool ):
                Default to None.
            output_hidden_states (bool ):
                Default to None.
            return_dict (bool ):
                Default to None.

        Returns:
            Union[Tuple, BaseModelOutputWithPooling]
        """

        output_attentions = (
            output_attentions if output_attentions is not None else
            self.config.output_attentions)
        output_hidden_states = (
            output_hidden_states if output_hidden_states is not None else
            self.config.output_hidden_states)
        return_dict = (
            return_dict
            if return_dict is not None else self.config.use_return_dict)

        input_shape = input_ids.size()
        input_ids = input_ids.view(-1, input_shape[-1])

        hidden_states = self.embeddings(input_ids)

        bsz, seq_len = input_shape
        causal_attention_mask = self._build_causal_attention_mask(
            bsz, seq_len, hidden_states.dtype).to(hidden_states.device)

        # expand attention_mask
        if attention_mask is not None:
            # [bsz, seq_len] -> [bsz, 1, tgt_seq_len, src_seq_len]
            attention_mask = _expand_mask(attention_mask, hidden_states.dtype)

        encoder_outputs = self.encoder(
            inputs_embeds=hidden_states,
            attention_mask=attention_mask,
            causal_attention_mask=causal_attention_mask,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )

        last_hidden_state = encoder_outputs[0]
        last_hidden_state = self.final_layer_norm(last_hidden_state)

        # text_embeds.shape = [batch_size, sequence_length, transformer.width]
        # take features from the eot embedding
        # (eot_token is the highest number in each sequence)
        # casting to torch.int for onnx compatibility:
        # argmax doesn't support int64 inputs with opset 14
        pooled_output = last_hidden_state[
            torch.arange(
                last_hidden_state.shape[0], device=last_hidden_state.device),
            input_ids.to(dtype=torch.int, device=last_hidden_state.device
                         ).argmax(dim=-1), ]

        if not return_dict:
            return (last_hidden_state, pooled_output) + encoder_outputs[1:]

        return BaseModelOutputWithPooling(
            last_hidden_state=last_hidden_state,
            pooler_output=pooled_output,
            hidden_states=encoder_outputs.hidden_states,
            attentions=encoder_outputs.attentions,
        )


class FastComposerCLIPImageEncoder(CLIPPreTrainedModel):
    """CLIPImageEncoder for FastComposerModel."""

    @staticmethod
    def from_pretrained(global_model_name_or_path):
        """Init CLIPModel with Clip model name or path."""

        model = CLIPModel.from_pretrained(global_model_name_or_path)
        vision_model = model.vision_model
        visual_projection = model.visual_projection
        vision_processor = T.Normalize(
            (0.48145466, 0.4578275, 0.40821073),
            (0.26862954, 0.26130258, 0.27577711),
        )
        return FastComposerCLIPImageEncoder(
            vision_model,
            visual_projection,
            vision_processor,
        )

    def __init__(
        self,
        vision_model,
        visual_projection,
        vision_processor,
    ):
        super().__init__(vision_model.config)
        self.vision_model = vision_model
        self.visual_projection = visual_projection
        self.vision_processor = vision_processor

        self.image_size = vision_model.config.image_size

    def forward(self, object_pixel_values):
        """Forward function.

        Args:
            object_pixel_values (torch.Tensor ):
                You can directly input a ``torch.Tensor``.

        Returns:
            torch.Tensor : ``torch.tensor`` will be returned.
        """

        b, num_objects, c, h, w = object_pixel_values.shape

        object_pixel_values = object_pixel_values.view(b * num_objects, c, h,
                                                       w)

        if h != self.image_size or w != self.image_size:
            h, w = self.image_size, self.image_size
            object_pixel_values = F.interpolate(
                object_pixel_values, (h, w), mode='bilinear')

        object_pixel_values = self.vision_processor(object_pixel_values)
        object_embeds = self.vision_model(object_pixel_values)[1]
        object_embeds = self.visual_projection(object_embeds)
        object_embeds = object_embeds.view(b, num_objects, 1, -1)
        return object_embeds


def get_object_transforms(cfg):
    """Get Object transforms."""

    if cfg['no_object_augmentation']:
        pre_augmentations = []
        augmentations = []
    else:
        pre_augmentations = [
            (
                'zoomin',
                T.RandomApply([RandomZoomIn(min_zoom=1.0, max_zoom=2.0)],
                              p=0.5),
            ),
        ]

        augmentations = [
            (
                'rotate',
                T.RandomApply(
                    [
                        T.RandomAffine(
                            degrees=30,
                            interpolation=T.InterpolationMode.BILINEAR)
                    ],
                    p=0.75,
                ),
            ),
            ('jitter', T.RandomApply([T.ColorJitter(0.5, 0.5, 0.5, 0.5)],
                                     p=0.5)),
            ('blur', T.RandomApply([T.GaussianBlur(5, sigma=(0.1, 2.0))],
                                   p=0.5)),
            ('gray', T.RandomGrayscale(p=0.1)),
            ('flip', T.RandomHorizontalFlip()),
            ('elastic', T.RandomApply([T.ElasticTransform()], p=0.5)),
        ]

    object_transforms = torch.nn.Sequential(
        OrderedDict([
            *pre_augmentations,
            ('pad_to_square', PadToSquare(fill=0, padding_mode='constant')),
            (
                'resize',
                T.Resize(
                    (cfg['object_resolution'], cfg['object_resolution']),
                    interpolation=T.InterpolationMode.BILINEAR,
                ),
            ),
            *augmentations,
            ('convert_to_float', T.ConvertImageDtype(torch.float32)),
        ]))
    return object_transforms


class FastComposerPostfuseModule(nn.Module):
    """Postfuse Module for FastComposerModel."""

    def __init__(self, embed_dim):
        super().__init__()
        self.mlp1 = MLP(
            embed_dim * 2, embed_dim, embed_dim, use_residual=False)
        self.mlp2 = MLP(embed_dim, embed_dim, embed_dim, use_residual=True)
        self.layer_norm = nn.LayerNorm(embed_dim)

    def fuse_fn(self, text_embeds, object_embeds):
        """Fuse function.

        Args:
            text_embeds (torch.Tensor ):
                You can directly input a ``torch.Tensor``.
            object_embeds (torch.Tensor ):
                You can directly input a ``torch.Tensor``.

        Returns:
            torch.Tensor : ``torch.tensor`` will be returned.
        """

        text_object_embeds = torch.cat([text_embeds, object_embeds], dim=-1)
        text_object_embeds = self.mlp1(text_object_embeds) + text_embeds
        text_object_embeds = self.mlp2(text_object_embeds)
        text_object_embeds = self.layer_norm(text_object_embeds)
        return text_object_embeds

    def forward(
        self,
        text_embeds,
        object_embeds,
        image_token_mask,
        num_objects,
    ) -> torch.Tensor:
        """Forward function.

        Args:
            text_embeds (torch.Tensor ):
                You can directly input a ``torch.Tensor``.
            object_embeds (torch.Tensor ):
                You can directly input a ``torch.Tensor``.
            image_token_mask (torch.Tensor ):
                You can directly input a ``torch.Tensor``.

        Returns:
            torch.Tensor : ``torch.tensor`` will be returned.
        """

        text_object_embeds = fuse_object_embeddings(text_embeds,
                                                    image_token_mask,
                                                    object_embeds, num_objects,
                                                    self.fuse_fn)

        return text_object_embeds


def unet_store_cross_attention_scores(unet, attention_scores, layers=5):
    """Unet store cross attention scores."""
    from diffusers.models.attention_processor import (Attention, AttnProcessor,
                                                      AttnProcessor2_0)

    UNET_LAYER_NAMES = [
        'down_blocks.0',
        'down_blocks.1',
        'down_blocks.2',
        'mid_block',
        'up_blocks.1',
        'up_blocks.2',
        'up_blocks.3',
    ]

    start_layer = (len(UNET_LAYER_NAMES) - layers) // 2
    end_layer = start_layer + layers
    applicable_layers = UNET_LAYER_NAMES[start_layer:end_layer]

    def make_new_get_attention_scores_fn(name):
        """Wrapper Function of create attention scores for unet."""

        def new_get_attention_scores(module, query, key, attention_mask=None):
            """Create attention scores for unet."""
            attention_probs = module.old_get_attention_scores(
                query, key, attention_mask)
            attention_scores[name] = attention_probs
            return attention_probs

        return new_get_attention_scores

    for name, module in unet.named_modules():
        if isinstance(module, Attention) and 'attn2' in name:
            if not any(layer in name for layer in applicable_layers):
                continue
            if isinstance(module.processor, AttnProcessor2_0):
                module.set_processor(AttnProcessor())
            module.old_get_attention_scores = module.get_attention_scores
            module.get_attention_scores = types.MethodType(
                make_new_get_attention_scores_fn(name), module)

    return unet


class BalancedL1Loss(nn.Module):
    """BalancedL1Loss for object localization."""

    def __init__(self, threshold=1.0, normalize=False):
        super().__init__()
        self.threshold = threshold
        self.normalize = normalize

    def forward(self, object_token_attn_prob, object_segmaps):
        """Forward function.

        Args:
            object_token_attn_prob (torch.Tensor ):
                You can directly input a ``torch.Tensor``.
            object_segmaps (torch.Tensor ):
                You can directly input a ``torch.Tensor``.

        Returns:
            float : ``float`` will be returned.
        """
        if self.normalize:
            object_token_attn_prob = object_token_attn_prob / (
                object_token_attn_prob.max(dim=2, keepdim=True)[0] + 1e-5)
        background_segmaps = 1 - object_segmaps
        background_segmaps_sum = background_segmaps.sum(dim=2) + 1e-5
        object_segmaps_sum = object_segmaps.sum(dim=2) + 1e-5

        background_loss = (object_token_attn_prob * background_segmaps).sum(
            dim=2) / background_segmaps_sum

        object_loss = (object_token_attn_prob *
                       object_segmaps).sum(dim=2) / object_segmaps_sum

        return background_loss - object_loss


def get_object_localization_loss(
    cross_attention_scores,
    object_segmaps,
    image_token_idx,
    image_token_idx_mask,
    loss_fn,
):
    """To obtain the average of the loss for each layer of object
    localization."""

    num_layers = len(cross_attention_scores)
    loss = 0
    for k, v in cross_attention_scores.items():
        layer_loss = get_object_localization_loss_for_one_layer(
            v, object_segmaps, image_token_idx, image_token_idx_mask, loss_fn)
        loss += layer_loss
    return loss / num_layers


def get_object_localization_loss_for_one_layer(
    cross_attention_scores,
    object_segmaps,
    object_token_idx,
    object_token_idx_mask,
    loss_fn,
):
    """Get object localization loss for one layer."""

    bxh, num_noise_latents, num_text_tokens = cross_attention_scores.shape
    b, max_num_objects, _, _ = object_segmaps.shape
    size = int(num_noise_latents**0.5)

    # Resize the object segmentation maps to
    # the size of the cross attention scores
    object_segmaps = F.interpolate(
        object_segmaps, size=(size, size), mode='bilinear')
    # (b, max_num_objects, size, size)

    object_segmaps = object_segmaps.view(
        b, max_num_objects, -1)  # (b, max_num_objects, num_noise_latents)

    num_heads = bxh // b

    cross_attention_scores = cross_attention_scores.view(
        b, num_heads, num_noise_latents, num_text_tokens)

    # Gather object_token_attn_prob
    object_token_attn_prob = torch.gather(
        cross_attention_scores,
        dim=3,
        index=object_token_idx.view(b, 1, 1, max_num_objects).expand(
            b, num_heads, num_noise_latents, max_num_objects),
    )  # (b, num_heads, num_noise_latents, max_num_objects)

    object_segmaps = (
        object_segmaps.permute(0, 2,
                               1).unsqueeze(1).expand(b, num_heads,
                                                      num_noise_latents,
                                                      max_num_objects))

    loss = loss_fn(object_token_attn_prob, object_segmaps)

    loss = loss * object_token_idx_mask.view(b, 1, max_num_objects)
    object_token_cnt = object_token_idx_mask.sum(dim=1).view(b, 1) + 1e-5
    loss = (loss.sum(dim=2) / object_token_cnt).mean()

    return loss


class RandomZoomIn(nn.Module):
    """RandomZoomIn for object transform."""

    def __init__(self, min_zoom=1.0, max_zoom=1.5):
        super().__init__()
        self.min_zoom = min_zoom
        self.max_zoom = max_zoom

    def forward(self, image: torch.Tensor):
        """Forward function.

        Args:
            image (torch.Tensor ): You can directly input a ``torch.Tensor``.

        Returns:
            torch.Tensor : ``torch.tensor`` will be returned.
        """
        zoom = torch.rand(1) * (self.max_zoom - self.min_zoom) + self.min_zoom
        image = T.functional.resize(
            image,
            (int(zoom * image.shape[1]), int(zoom * image.shape[2])),
            interpolation=T.InterpolationMode.BILINEAR,
        )
        # crop top square
        image = CropTopSquare()(image)
        return image


class PadToSquare(nn.Module):
    """If the height of the image is greater than the width, padding will be
    added on both sides of the image to make it a square."""

    def __init__(self, fill=0, padding_mode='constant'):
        super().__init__()
        self.fill = fill
        self.padding_mode = padding_mode

    def forward(self, image: torch.Tensor):
        """Forward function.

        Args:
            image (torch.Tensor ): You can directly input a ``torch.Tensor``.

        Returns:
            torch.Tensor : ``torch.tensor`` will be returned.
        """
        _, h, w = image.shape
        if h == w:
            return image
        elif h > w:
            padding = (h - w) // 2
            image = torch.nn.functional.pad(
                image,
                (padding, padding, 0, 0),
                self.padding_mode,
                self.fill,
            )
        else:
            padding = (w - h) // 2
            image = torch.nn.functional.pad(
                image,
                (0, 0, padding, padding),
                self.padding_mode,
                self.fill,
            )
        return image


class CropTopSquare(nn.Module):
    """If the height of the image is greater than the width, the image will be
    cropped into a square starting from the top of the image."""

    def __init__(self):
        super().__init__()

    def forward(self, image: torch.Tensor):
        """Forward function.

        Args:
            image (torch.Tensor ): You can directly input a ``torch.Tensor``.

        Returns:
            torch.Tensor : ``torch.tensor`` will be returned.
        """
        _, h, w = image.shape
        if h <= w:
            return image
        return image[:, :w, :]


class MLP(nn.Module):
    """Multilayer Perceptron."""

    def __init__(self, in_dim, out_dim, hidden_dim, use_residual=True):
        super().__init__()
        if use_residual:
            assert in_dim == out_dim
        self.layernorm = nn.LayerNorm(in_dim)
        self.fc1 = nn.Linear(in_dim, hidden_dim)
        self.fc2 = nn.Linear(hidden_dim, out_dim)
        self.use_residual = use_residual
        self.act_fn = nn.GELU()

    def forward(self, x):
        """Forward function.

        Args:
            x (torch.Tensor ): You can directly input a ``torch.Tensor``.

        Returns:
            torch.Tensor : ``torch.tensor`` will be returned.
        """
        residual = x
        x = self.layernorm(x)
        x = self.fc1(x)
        x = self.act_fn(x)
        x = self.fc2(x)
        if self.use_residual:
            x = x + residual
        return x


def fuse_object_embeddings(
    inputs_embeds,
    image_token_mask,
    object_embeds,
    num_objects,
    fuse_fn=torch.add,
):
    """Fuse object embeddings."""

    object_embeds = object_embeds.to(inputs_embeds.dtype)

    batch_size, max_num_objects = object_embeds.shape[:2]
    seq_length = inputs_embeds.shape[1]
    flat_object_embeds = object_embeds.view(-1, object_embeds.shape[-2],
                                            object_embeds.shape[-1])

    valid_object_mask = (
        torch.arange(max_num_objects,
                     device=flat_object_embeds.device)[None, :] <
        num_objects[:, None])

    valid_object_embeds = flat_object_embeds[valid_object_mask.flatten()]

    inputs_embeds = inputs_embeds.view(-1, inputs_embeds.shape[-1])
    image_token_mask = image_token_mask.view(-1)
    valid_object_embeds = valid_object_embeds.view(
        -1, valid_object_embeds.shape[-1])

    # slice out the image token embeddings
    image_token_embeds = inputs_embeds[image_token_mask]
    valid_object_embeds = fuse_fn(image_token_embeds, valid_object_embeds)

    inputs_embeds.masked_scatter_(image_token_mask[:, None],
                                  valid_object_embeds)
    inputs_embeds = inputs_embeds.view(batch_size, seq_length, -1)
    return inputs_embeds


def build_causal_attention_mask(bsz, seq_len, dtype, device=None):
    """The function originally belonged to CLIPTextTransformer, but it has been
    removed in versions of transformers after 4.25.1."""

    # lazily create causal attention mask,
    # with full attention between the vision tokens
    # pytorch uses additive attention mask; fill with -inf
    mask = torch.empty(bsz, seq_len, seq_len, dtype=dtype, device=device)
    mask.fill_(torch.tensor(torch.finfo(dtype).min))
    mask.triu_(1)  # zero out the lower diagonal
    mask = mask.unsqueeze(1)  # expand mask
    return mask



# Copyright (c) OpenMMLab. All rights reserved.
import os
from typing import Dict, List

import numpy as np
from mmengine import mkdir_or_exist
from PIL.Image import Image
from torchvision.utils import save_image

from .base_mmagic_inferencer import BaseMMagicInferencer, InputsType, PredType


class DiffusersPipelineInferencer(BaseMMagicInferencer):
    """inferencer that predicts with text2image models."""

    func_kwargs = dict(
        preprocess=[
            'text', 'negative_prompt', 'num_inference_steps', 'height', 'width'
        ],
        forward=[],
        visualize=['result_out_dir'],
        postprocess=[])

    def preprocess(self,
                   text: InputsType = None,
                   negative_prompt: InputsType = None,
                   num_inference_steps: int = 20,
                   height=None,
                   width=None) -> Dict:
        """Process the inputs into a model-feedable format.

        Args:
            text(InputsType): text input for text-to-image model.
            negative_prompt(InputsType): negative prompt.

        Returns:
            result(Dict): Results of preprocess.
        """
        result = self.extra_parameters
        if text:
            result['prompt'] = text
        if negative_prompt:
            result['negative_prompt'] = negative_prompt
        if num_inference_steps:
            result['num_inference_steps'] = num_inference_steps
        if height:
            result['height'] = height
        if width:
            result['width'] = width

        return result

    def forward(self, inputs: InputsType) -> PredType:
        """Forward the inputs to the model."""
        images = self.model(**inputs).images

        return images

    def visualize(self,
                  preds: PredType,
                  result_out_dir: str = None) -> List[np.ndarray]:
        """Visualize predictions.

        Args:
            preds (List[Union[str, np.ndarray]]): Forward results
                by the inferencer.
            result_out_dir (str): Output directory of image.
                Defaults to ''.

        Returns:
            List[np.ndarray]: Result of visualize
        """
        if result_out_dir:
            mkdir_or_exist(os.path.dirname(result_out_dir))
            if type(preds) is list:
                preds = preds[0]
            if type(preds) is Image:
                preds.save(result_out_dir)
            else:
                save_image(preds, result_out_dir, normalize=True)

        return preds



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, List, Tuple

import torch
import torch.nn.functional as F
from mmengine.optim import OptimWrapper
from torch import Tensor

from mmagic.registry import MODELS
from mmagic.structures import DataSample
from ...base_models import BaseGAN


@MODELS.register_module()
class LSGAN(BaseGAN):
    """Implementation of `Least Squares Generative Adversarial Networks`.

    Paper link: https://arxiv.org/pdf/1611.04076.pdf

    Detailed architecture can be found in
    :class:`~mmagic.models.editors.lsgan.LSGANGenerator`
    and
    :class:`~mmagic.models.editors.lsgan.LSGANDiscriminator`
    """

    def disc_loss(self, disc_pred_fake: Tensor,
                  disc_pred_real: Tensor) -> Tuple:
        r"""Get disc loss. LSGAN use the least squares loss to train
        the discriminator.

        .. math::
            L_{D}=\left(D\left(X_{\text {data }}\right)-1\right)^{2}
                +(D(G(z)))^{2}


        Args:
            disc_pred_fake (Tensor): Discriminator's prediction of the fake
                images.
            disc_pred_real (Tensor): Discriminator's prediction of the real
                images.

        Returns:
            tuple[Tensor, dict]: Loss value and a dict of log variables.
        """
        losses_dict = dict()
        losses_dict['loss_disc_fake'] = F.mse_loss(
            disc_pred_fake, 0. * torch.ones_like(disc_pred_fake))
        losses_dict['loss_disc_real'] = F.mse_loss(
            disc_pred_real, 1. * torch.ones_like(disc_pred_real))

        loss, log_var = self.parse_losses(losses_dict)
        return loss, log_var

    def gen_loss(self, disc_pred_fake: Tensor) -> Tuple:
        """Get gen loss. LSGAN use the least squares loss to train the
        generator.

        .. math::
            L_{G}=(D(G(z))-1)^{2}

        Args:
            disc_pred_fake (Tensor): Discriminator's prediction of the fake
                images.

        Returns:
            tuple[Tensor, dict]: Loss value and a dict of log variables.
        """
        losses_dict = dict()
        losses_dict['loss_gen'] = F.mse_loss(
            disc_pred_fake, 1. * torch.ones_like(disc_pred_fake))
        loss, log_var = self.parse_losses(losses_dict)
        return loss, log_var

    def train_discriminator(self, inputs: dict, data_samples: DataSample,
                            optimizer_wrapper: OptimWrapper
                            ) -> Dict[str, Tensor]:
        """Train discriminator.

        Args:
            inputs (dict): Inputs from dataloader.
            data_samples (DataSample): Data samples from dataloader.
            optim_wrapper (OptimWrapper): OptimWrapper instance used to update
                model parameters.
        Returns:
            Dict[str, Tensor]: A ``dict`` of tensor for logging.
        """
        real_imgs = data_samples.gt_img

        num_batches = real_imgs.shape[0]

        noise_batch = self.noise_fn(num_batches=num_batches)
        with torch.no_grad():
            fake_imgs = self.generator(noise=noise_batch, return_noise=False)

        disc_pred_fake = self.discriminator(fake_imgs)
        disc_pred_real = self.discriminator(real_imgs)

        parsed_losses, log_vars = self.disc_loss(disc_pred_fake,
                                                 disc_pred_real)
        optimizer_wrapper.update_params(parsed_losses)
        return log_vars

    def train_generator(self, inputs: dict, data_samples: List[DataSample],
                        optimizer_wrapper: OptimWrapper) -> Dict[str, Tensor]:
        """Train generator.

        Args:
            inputs (dict): Inputs from dataloader.
            data_samples (List[DataSample]): Data samples from dataloader.
                Do not used in generator's training.
            optim_wrapper (OptimWrapper): OptimWrapper instance used to update
                model parameters.

        Returns:
            Dict[str, Tensor]: A ``dict`` of tensor for logging.
        """
        num_batches = len(data_samples)

        noise = self.noise_fn(num_batches=num_batches)
        fake_imgs = self.generator(noise=noise, return_noise=False)

        disc_pred_fake = self.discriminator(fake_imgs)
        parsed_loss, log_vars = self.gen_loss(disc_pred_fake)

        optimizer_wrapper.update_params(parsed_loss)
        return log_vars



# Copyright (c) OpenMMLab. All rights reserved.
import torch


def get_unknown_tensor(trimap, unknown_value=128 / 255):
    """Get 1-channel unknown area tensor from the 3 or 1-channel trimap tensor.

    Args:
        trimap (Tensor): Tensor with shape (N, 3, H, W) or (N, 1, H, W).
        unknown_value (float): Scalar value indicating unknown region in
            trimap.
            If trimap is pre-processed using `'rescale_to_zero_one'`, then
            0 for bg, 128/255 for unknown, 1 for fg,
            and unknown_value should set to 128 / 255.
            If trimap is pre-processed by
            :meth:`FormatTrimap(to_onehot=False)`, then
            0 for bg, 1 for unknown, 2 for fg
            and unknown_value should set to 1.
            If trimap is pre-processed by
            :meth:`FormatTrimap(to_onehot=True)`, then
            trimap is 3-channeled, and this value is not used.

    Returns:
        Tensor: Unknown area mask of shape (N, 1, H, W).
    """
    if trimap.shape[1] == 3:
        # The three channels correspond to (bg mask, unknown mask, fg mask)
        # respectively.
        weight = trimap[:, 1:2, :, :].float()
    # elif 'to_onehot' in meta[0]:
    # key 'to_onehot' is added by pipeline `FormatTrimap`
    # 0 for bg, 1 for unknown, 2 for fg
    # weight = trimap.eq(1).float()
    else:
        # trimap is simply processed by pipeline `RescaleToZeroOne`
        # 0 for bg, 128/255 for unknown, 1 for fg
        weight = trimap.eq(unknown_value).float()
    return weight


def normalize_vecs(vectors: torch.Tensor) -> torch.Tensor:
    """Normalize vector with it's lengths at the last dimension. If `vector` is
    two-dimension tensor, this function is same as L2 normalization.

    Args:
        vector (torch.Tensor): Vectors to be normalized.

    Returns:
        torch.Tensor: Vectors after normalization.
    """
    return vectors / (torch.norm(vectors, dim=-1, keepdim=True))



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Optional

import torch
import torch.nn as nn

from mmagic.registry import MODELS
from ..gan_loss import (disc_shift_loss, gradient_penalty_loss,
                        r1_gradient_penalty_loss)


@MODELS.register_module()
class DiscShiftLossComps(nn.Module):
    """Disc Shift Loss.

    This loss is proposed in PGGAN as an auxiliary loss for discriminator.

    **Note for the design of ``data_info``:**
    In ``MMagic``, almost all of loss modules contain the argument
    ``data_info``, which can be used for constructing the link between the
    input items (needed in loss calculation) and the data from the generative
    model. For example, in the training of GAN model, we will collect all of
    important data/modules into a dictionary:

    .. code-block:: python
        :caption: Code from StaticUnconditionalGAN, train_step
        :linenos:

        data_dict_ = dict(
            gen=self.generator,
            disc=self.discriminator,
            disc_pred_fake=disc_pred_fake,
            disc_pred_real=disc_pred_real,
            fake_imgs=fake_imgs,
            real_imgs=real_imgs,
            iteration=curr_iter,
            batch_size=batch_size)

    But in this loss, we will need to provide ``pred`` as input. Thus, an
    example of the ``data_info`` is:

    .. code-block:: python
        :linenos:

        data_info = dict(
            pred='disc_pred_fake')

    Then, the module will automatically construct this mapping from the input
    data dictionary.

    In addition, in general, ``disc_shift_loss`` will be applied over real and
    fake data. In this case, users just need to add this loss module twice, but
    with different ``data_info``. Our model will automatically add these two
    items.

    Args:
        loss_weight (float, optional): Weight of this loss item.
            Defaults to ``1.``.
        data_info (dict, optional): Dictionary contains the mapping between
            loss input args and data dictionary. If ``None``, this module will
            directly pass the input data to the loss function.
            Defaults to None.
        loss_name (str, optional): Name of the loss item. If you want this loss
            item to be included into the backward graph, `loss_` must be the
            prefix of the name. Defaults to 'loss_disc_shift'.
    """

    def __init__(self,
                 loss_weight: float = 1.0,
                 data_info: Optional[dict] = None,
                 loss_name: str = 'loss_disc_shift') -> None:
        super().__init__()
        self.loss_weight = loss_weight
        self.data_info = data_info
        self._loss_name = loss_name

    def forward(self, *args, **kwargs) -> torch.Tensor:
        """Forward function.

        If ``self.data_info`` is not ``None``, a dictionary containing all of
        the data and necessary modules should be passed into this function.
        If this dictionary is given as a non-keyword argument, it should be
        offered as the first argument. If you are using keyword argument,
        please name it as `outputs_dict`.

        If ``self.data_info`` is ``None``, the input argument or key-word
        argument will be directly passed to loss function, ``disc_shift_loss``.
        """
        # use data_info to build computational path
        if self.data_info is not None:
            # parse the args and kwargs
            if len(args) == 1:
                assert isinstance(args[0], dict), (
                    'You should offer a dictionary containing network outputs '
                    'for building up computational graph of this loss module.')
                outputs_dict = args[0]
            elif 'outputs_dict' in kwargs:
                assert len(args) == 0, (
                    'If the outputs dict is given in keyworded arguments, no'
                    ' further non-keyworded arguments should be offered.')
                outputs_dict = kwargs.pop('outputs_dict')
            else:
                raise NotImplementedError(
                    'Cannot parsing your arguments passed to this loss module.'
                    ' Please check the usage of this module')
            # link the outputs with loss input args according to self.data_info
            loss_input_dict = {
                k: outputs_dict[v]
                for k, v in self.data_info.items()
            }
            kwargs.update(loss_input_dict)
            return disc_shift_loss(**kwargs) * self.loss_weight
        else:
            # if you have not define how to build computational graph, this
            # module will just directly return the loss as usual.
            return disc_shift_loss(*args, **kwargs) * self.loss_weight

    def loss_name(self) -> str:
        """Loss Name.

        This function must be implemented and will return the name of this
        loss function. This name will be used to combine different loss items
        by simple sum operation. In addition, if you want this loss item to be
        included into the backward graph, `loss_` must be the prefix of the
        name.

        Returns:
            str: The name of this loss item.
        """
        return self._loss_name


@MODELS.register_module()
class GradientPenaltyLossComps(nn.Module):
    """Gradient Penalty for WGAN-GP.

    In the detailed implementation, there are two streams where one uses the
    pixel-wise gradient norm, but the other adopts normalization along instance
    (HWC) dimensions. Thus, ``norm_mode`` are offered to define which mode you
    want.

    **Note for the design of ``data_info``:**
    In ``MMagic``, almost all of loss modules contain the argument
    ``data_info``, which can be used for constructing the link between the
    input items (needed in loss calculation) and the data from the generative
    model. For example, in the training of GAN model, we will collect all of
    important data/modules into a dictionary:

    .. code-block:: python
        :caption: Code from StaticUnconditionalGAN, train_step
        :linenos:

        data_dict_ = dict(
            gen=self.generator,
            disc=self.discriminator,
            disc_pred_fake=disc_pred_fake,
            disc_pred_real=disc_pred_real,
            fake_imgs=fake_imgs,
            real_imgs=real_imgs,
            iteration=curr_iter,
            batch_size=batch_size)

    But in this loss, we will need to provide ``discriminator``, ``real_data``,
    and ``fake_data`` as input. Thus, an example of the ``data_info`` is:

    .. code-block:: python
        :linenos:

        data_info = dict(
            discriminator='disc',
            real_data='real_imgs',
            fake_data='fake_imgs')

    Then, the module will automatically construct this mapping from the input
    data dictionary.

    Args:
        loss_weight (float, optional): Weight of this loss item.
            Defaults to ``1.``.
        data_info (dict, optional): Dictionary contains the mapping between
            loss input args and data dictionary. If ``None``, this module will
            directly pass the input data to the loss function.
            Defaults to None.
        norm_mode (str): This argument decides along which dimension the norm
            of the gradients will be calculated. Currently, we support ["pixel"
            , "HWC"]. Defaults to "pixel".
        loss_name (str, optional): Name of the loss item. If you want this loss
            item to be included into the backward graph, `loss_` must be the
            prefix of the name. Defaults to 'loss_gp'.
    """

    def __init__(self,
                 loss_weight: float = 1.0,
                 norm_mode: str = 'pixel',
                 data_info: Optional[dict] = None,
                 loss_name: str = 'loss_gp') -> None:
        super().__init__()
        self.loss_weight = loss_weight
        self.norm_mode = norm_mode
        self.data_info = data_info
        self._loss_name = loss_name

    def forward(self, *args, **kwargs) -> torch.Tensor:
        """Forward function.

        If ``self.data_info`` is not ``None``, a dictionary containing all of
        the data and necessary modules should be passed into this function.
        If this dictionary is given as a non-keyword argument, it should be
        offered as the first argument. If you are using keyword argument,
        please name it as `outputs_dict`.

        If ``self.data_info`` is ``None``, the input argument or key-word
        argument will be directly passed to loss function,
        ``gradient_penalty_loss``.
        """
        # use data_info to build computational path
        if self.data_info is not None:
            # parse the args and kwargs
            if len(args) == 1:
                assert isinstance(args[0], dict), (
                    'You should offer a dictionary containing network outputs '
                    'for building up computational graph of this loss module.')
                outputs_dict = args[0]
            elif 'outputs_dict' in kwargs:
                assert len(args) == 0, (
                    'If the outputs dict is given in keyworded arguments, no'
                    ' further non-keyworded arguments should be offered.')
                outputs_dict = kwargs.pop('outputs_dict')
            else:
                raise NotImplementedError(
                    'Cannot parsing your arguments passed to this loss module.'
                    ' Please check the usage of this module')
            # link the outputs with loss input args according to self.data_info
            loss_input_dict = {
                k: outputs_dict[v]
                for k, v in self.data_info.items()
            }
            kwargs.update(loss_input_dict)
            # kwargs.update(
            #     dict(weight=self.loss_weight, norm_mode=self.norm_mode))
            kwargs.update(dict(norm_mode=self.norm_mode))
            return gradient_penalty_loss(**kwargs) * self.loss_weight
        else:
            # if you have not define how to build computational graph, this
            # module will just directly return the loss as usual.
            # return gradient_penalty_loss(
            #     *args, weight=self.loss_weight, **kwargs)
            return gradient_penalty_loss(*args, **kwargs) * self.loss_weight

    def loss_name(self) -> str:
        """Loss Name.

        This function must be implemented and will return the name of this
        loss function. This name will be used to combine different loss items
        by simple sum operation. In addition, if you want this loss item to be
        included into the backward graph, `loss_` must be the prefix of the
        name.

        Returns:
            str: The name of this loss item.
        """
        return self._loss_name


@MODELS.register_module()
class R1GradientPenaltyComps(nn.Module):
    """R1 gradient penalty for WGAN-GP.

    R1 regularizer comes from:
    "Which Training Methods for GANs do actually Converge?" ICML'2018

    Different from original gradient penalty, this regularizer only penalized
    gradient w.r.t. real data.

    **Note for the design of ``data_info``:**
    In ``MMagic``, almost all of loss modules contain the argument
    ``data_info``, which can be used for constructing the link between the
    input items (needed in loss calculation) and the data from the generative
    model. For example, in the training of GAN model, we will collect all of
    important data/modules into a dictionary:

    .. code-block:: python
        :caption: Code from StaticUnconditionalGAN, train_step
        :linenos:

        data_dict_ = dict(
            gen=self.generator,
            disc=self.discriminator,
            disc_pred_fake=disc_pred_fake,
            disc_pred_real=disc_pred_real,
            fake_imgs=fake_imgs,
            real_imgs=real_imgs,
            iteration=curr_iter,
            batch_size=batch_size)

    But in this loss, we will need to provide ``discriminator`` and
    ``real_data`` as input. Thus, an example of the ``data_info`` is:

    .. code-block:: python
        :linenos:

        data_info = dict(
            discriminator='disc',
            real_data='real_imgs')

    Then, the module will automatically construct this mapping from the input
    data dictionary.

    Args:
        loss_weight (float, optional): Weight of this loss item.
            Defaults to ``1.``.
        data_info (dict, optional): Dictionary contains the mapping between
            loss input args and data dictionary. If ``None``, this module will
            directly pass the input data to the loss function.
            Defaults to None.
        norm_mode (str): This argument decides along which dimension the norm
            of the gradients will be calculated. Currently, we support ["pixel"
            , "HWC"]. Defaults to "pixel".
        interval (int, optional): The interval of calculating this loss.
            Defaults to 1.
        loss_name (str, optional): Name of the loss item. If you want this loss
            item to be included into the backward graph, `loss_` must be the
            prefix of the name. Defaults to 'loss_r1_gp'.
    """

    def __init__(self,
                 loss_weight: float = 1.0,
                 norm_mode: str = 'pixel',
                 interval: int = 1,
                 data_info: Optional[dict] = None,
                 use_apex_amp: bool = False,
                 loss_name: str = 'loss_r1_gp') -> None:
        super().__init__()
        self.loss_weight = loss_weight
        self.norm_mode = norm_mode
        self.interval = interval
        self.data_info = data_info
        self.use_apex_amp = use_apex_amp
        self._loss_name = loss_name

    def forward(self, *args, **kwargs) -> torch.Tensor:
        """Forward function.

        If ``self.data_info`` is not ``None``, a dictionary containing all of
        the data and necessary modules should be passed into this function.
        If this dictionary is given as a non-keyword argument, it should be
        offered as the first argument. If you are using keyword argument,
        please name it as `outputs_dict`.

        If ``self.data_info`` is ``None``, the input argument or key-word
        argument will be directly passed to loss function,
        ``r1_gradient_penalty_loss``.
        """
        if self.interval > 1:
            assert self.data_info is not None
        # use data_info to build computational path
        if self.data_info is not None:
            # parse the args and kwargs
            if len(args) == 1:
                assert isinstance(args[0], dict), (
                    'You should offer a dictionary containing network outputs '
                    'for building up computational graph of this loss module.')
                outputs_dict = args[0]
            elif 'outputs_dict' in kwargs:
                assert len(args) == 0, (
                    'If the outputs dict is given in keyworded arguments, no'
                    ' further non-keyworded arguments should be offered.')
                outputs_dict = kwargs.pop('outputs_dict')
            else:
                raise NotImplementedError(
                    'Cannot parsing your arguments passed to this loss module.'
                    ' Please check the usage of this module')
            if self.interval > 1 and outputs_dict[
                    'iteration'] % self.interval != 0:
                return None
            # link the outputs with loss input args according to self.data_info
            loss_input_dict = {
                k: outputs_dict[v]
                for k, v in self.data_info.items()
            }
            kwargs.update(loss_input_dict)
            kwargs.update(
                dict(norm_mode=self.norm_mode, use_apex_amp=self.use_apex_amp))
            return r1_gradient_penalty_loss(**kwargs) * self.loss_weight
        else:
            # if you have not define how to build computational graph, this
            # module will just directly return the loss as usual.
            return r1_gradient_penalty_loss(
                *args, norm_mode=self.norm_mode, **kwargs) * self.loss_weight

    def loss_name(self) -> str:
        """Loss Name.

        This function must be implemented and will return the name of this
        loss function. This name will be used to combine different loss items
        by simple sum operation. In addition, if you want this loss item to be
        included into the backward graph, `loss_` must be the prefix of the
        name.

        Returns:
            str: The name of this loss item.
        """
        return self._loss_name



_base_ = [
    '../_base_/datasets/ffhq_flip.py', '../_base_/models/base_styleganv2.py',
    '../_base_/gen_default_runtime.py'
]

ema_half_life = 10.
ema_config = dict(
    type='ExponentialMovingAverage',
    interval=1,
    momentum=1. - (0.5**(32. / (ema_half_life * 1000.))))

model = dict(
    type='MSPIEStyleGAN2',
    generator=dict(
        type='MSStyleGANv2Generator',
        head_pos_encoding=dict(
            type='SPE',
            embedding_dim=256,
            padding_idx=0,
            init_size=256,
            center_shift=100),
        deconv2conv=True,
        up_after_conv=True,
        up_config=dict(scale_factor=2, mode='bilinear', align_corners=True),
        out_size=256),
    discriminator=dict(
        type='MSStyleGAN2Discriminator', in_size=256, with_adaptive_pool=True),
    ema_config=ema_config,
    train_settings=dict(
        num_upblocks=6,
        multi_input_scales=[0, 2, 4],
        multi_scale_probability=[0.5, 0.25, 0.25]))

train_cfg = dict(max_iters=1100002)

# `batch_size` and `data_root` need to be set.
batch_size = 3
data_root = './data/ffhq/ffhq_imgs/ffhq_512'
train_dataloader = dict(
    batch_size=batch_size,  # set by user
    dataset=dict(data_root=data_root))

pipeline = [
    dict(type='LoadImageFromFile', key='img'),
    dict(type='Resize', scale=(256, 256)),
    dict(type='PackInputs', keys=['img'])
]

val_dataloader = dict(
    batch_size=batch_size,  # set by user
    dataset=dict(data_root=data_root, pipeline=pipeline))

test_dataloader = dict(
    batch_size=batch_size,  # set by user
    dataset=dict(data_root=data_root, pipeline=pipeline))

# define optimizer
d_reg_interval = 16
g_reg_interval = 4

g_reg_ratio = g_reg_interval / (g_reg_interval + 1)
d_reg_ratio = d_reg_interval / (d_reg_interval + 1)

optim_wrapper = dict(
    generator=dict(
        optimizer=dict(
            type='Adam', lr=0.002 * g_reg_ratio, betas=(0,
                                                        0.99**g_reg_ratio))),
    discriminator=dict(
        optimizer=dict(
            type='Adam', lr=0.002 * d_reg_ratio, betas=(0,
                                                        0.99**d_reg_ratio))))

# VIS_HOOK
custom_hooks = [
    dict(
        type='VisualizationHook',
        interval=5000,
        fixed_input=True,
        vis_kwargs_list=dict(type='GAN', name='fake_img'))
]
default_hooks = dict(checkpoint=dict(save_best=['FID-50k/fid'], rule=['less']))
# METRICS
metrics = [
    dict(
        type='FrechetInceptionDistance',
        prefix='FID-50k',
        fake_nums=50000,
        real_nums=50000,
        inception_style='pytorch',
        sample_model='ema'),
    dict(type='PrecisionAndRecall', fake_nums=10000, prefix='PR-10K')
]

val_evaluator = dict(metrics=metrics)
test_evaluator = dict(metrics=metrics)



_base_ = ['swinir_psnr-x2s64w8d6e180_8xb4-lr1e-4-600k_df2k-ost.py']

experiment_name = 'swinir_gan-x2s64w8d6e180_8xb4-lr1e-4-600k_df2k-ost.py'
work_dir = f'./work_dirs/{experiment_name}'
save_dir = './work_dirs/'



# Copyright (c) OpenMMLab. All rights reserved.
# Copyright (c) 2021, NVIDIA CORPORATION & AFFILIATES.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import sys
import traceback

import matplotlib.cm
import matplotlib.font_manager
import numpy as np
import torch
import torch.fft
import torch.nn.functional as F
from PIL import Image, ImageDraw, ImageFont

from demo.utils.gradio_utils import EasyDict
from mmagic.apis import MMagicInferencer


class CapturedException(Exception):

    def __init__(self, msg=None):
        if msg is None:
            _type, value, _traceback = sys.exc_info()
            assert value is not None
            if isinstance(value, CapturedException):
                msg = str(value)
            else:
                msg = traceback.format_exc()
        assert isinstance(msg, str)
        super().__init__(msg)


'''
----------------------------------------------------------------------------
'''


class CaptureSuccess(Exception):

    def __init__(self, out):
        super().__init__()
        self.out = out


'''
----------------------------------------------------------------------------
'''


def add_watermark_np(input_image_array, watermark_text='AI Generated'):
    image = Image.fromarray(np.uint8(input_image_array)).convert('RGBA')

    # Initialize text image
    txt = Image.new('RGBA', image.size, (255, 255, 255, 0))
    ttf_pth = matplotlib.font_manager.findSystemFonts(
        fontpaths=None, fontext='ttf')[-1]

    font = ImageFont.truetype(ttf_pth, round(25 / 512 * image.size[0]))
    d = ImageDraw.Draw(txt)

    text_width, text_height = font.getsize(watermark_text)
    text_position = (image.size[0] - text_width - 10,
                     image.size[1] - text_height - 10)
    text_color = (
        255, 255, 255, 128
    )  # white color with the alpha channel set to semi-transparent

    # Draw the text onto the text canvas
    d.text(text_position, watermark_text, font=font, fill=text_color)

    # Combine the image with the watermark
    watermarked = Image.alpha_composite(image, txt)
    watermarked_array = np.array(watermarked)
    return watermarked_array


class Renderer:

    def __init__(self, disable_timing=False):
        self._device = torch.device('cuda' if torch.cuda.is_available(
        ) else 'mps' if torch.backends.mps.is_available() else 'cpu')
        self._dtype = torch.float32 \
            if self._device.type == 'mps' else torch.float64
        self._pkl_data = dict()  # {pkl: dict | CapturedException, ...}
        self._networks = dict()  # {cache_key: torch.nn.Module, ...}
        self._pinned_bufs = dict()  # {(shape, dtype): torch.Tensor, ...}
        self._cmaps = dict()  # {name: torch.Tensor, ...}
        self._is_timing = False
        if not disable_timing:
            self._start_event = torch.cuda.Event(enable_timing=True)
            self._end_event = torch.cuda.Event(enable_timing=True)
        self._disable_timing = disable_timing
        self._net_layers = dict()  # {cache_key: [dnnlib.EasyDict, ...], ...}

    def render(self, **args):
        if self._disable_timing:
            self._is_timing = False
        else:
            self._start_event.record(torch.cuda.current_stream(self._device))
            self._is_timing = True
        res = EasyDict()
        try:
            init_net = False
            if not hasattr(self, 'G'):
                init_net = True
            if hasattr(self, 'pkl'):
                if self.pkl != args['pkl']:
                    init_net = True
            if hasattr(self, 'w_load'):
                if self.w_load is not args['w_load']:
                    init_net = True
            if hasattr(self, 'w0_seed'):
                if self.w0_seed != args['w0_seed']:
                    init_net = True
            if hasattr(self, 'w_plus'):
                if self.w_plus != args['w_plus']:
                    init_net = True
            if args['reset_w']:
                init_net = True
            res.init_net = init_net
            if init_net:
                self.init_network(res, **args)
            self._render_drag_impl(res, **args)
        except CapturedException:
            res.error = CapturedException()
        if not self._disable_timing:
            self._end_event.record(torch.cuda.current_stream(self._device))
        if 'image' in res:
            res.image = self.to_cpu(res.image).detach().numpy()
            res.image = add_watermark_np(res.image, 'AI Generated')
        if 'stats' in res:
            res.stats = self.to_cpu(res.stats).detach().numpy()
        if 'error' in res:
            res.error = str(res.error)
        # if 'stop' in res and res.stop:

        if self._is_timing and not self._disable_timing:
            self._end_event.synchronize()
            res.render_time = self._start_event.elapsed_time(
                self._end_event) * 1e-3
            self._is_timing = False
        return res

    def get_network(self, pkl, key, **tweak_kwargs):
        data = self._pkl_data.get(pkl, None)
        if data is None:
            print(f'Loading "{pkl}"... ', end='', flush=True)
            try:
                print('Done.')
            except CapturedException:
                data = CapturedException()
                print('Failed!')
            self._pkl_data[pkl] = data
            self._ignore_timing()
        if isinstance(data, CapturedException):
            raise data

        orig_net = data[key]
        cache_key = (orig_net, self._device,
                     tuple(sorted(tweak_kwargs.items())))
        net = self._networks.get(cache_key, None)
        if net is None:
            try:
                if 'stylegan2' in pkl:
                    from training.networks_stylegan2 import Generator
                else:
                    raise NameError('Cannot infer model type from pkl name!')

                print(data[key].init_args)
                print(data[key].init_kwargs)
                if 'stylegan_human' in pkl:
                    net = Generator(
                        *data[key].init_args,
                        **data[key].init_kwargs,
                        square=False,
                        padding=True)
                else:
                    net = Generator(*data[key].init_args,
                                    **data[key].init_kwargs)
                net.load_state_dict(data[key].state_dict())
                net.to(self._device)
            except CapturedException:
                net = CapturedException()
            self._networks[cache_key] = net
            self._ignore_timing()
        if isinstance(net, CapturedException):
            raise net
        return net

    def _get_pinned_buf(self, ref):
        key = (tuple(ref.shape), ref.dtype)
        buf = self._pinned_bufs.get(key, None)
        if buf is None:
            buf = torch.empty(ref.shape, dtype=ref.dtype).pin_memory()
            self._pinned_bufs[key] = buf
        return buf

    def to_device(self, buf):
        return self._get_pinned_buf(buf).copy_(buf).to(self._device)

    def to_cpu(self, buf):
        return self._get_pinned_buf(buf).copy_(buf).clone()

    def _ignore_timing(self):
        self._is_timing = False

    def _apply_cmap(self, x, name='viridis'):
        cmap = self._cmaps.get(name, None)
        if cmap is None:
            cmap = matplotlib.cm.get_cmap(name)
            cmap = cmap(np.linspace(0, 1, num=1024), bytes=True)[:, :3]
            cmap = self.to_device(torch.from_numpy(cmap))
            self._cmaps[name] = cmap
        hi = cmap.shape[0] - 1
        x = (x * hi + 0.5).clamp(0, hi).to(torch.int64)
        x = torch.nn.functional.embedding(x, cmap)
        return x

    def init_network(self,
                     res,
                     ckpt_pth=None,
                     w0_seed=0,
                     w_load=None,
                     w_plus=True,
                     noise_mode='const',
                     trunc_psi=0.7,
                     trunc_cutoff=None,
                     input_transform=None,
                     lr=0.001,
                     **kwargs):
        # Dig up network details.
        if '256' in ckpt_pth:
            editor = MMagicInferencer(
                'draggan',
                model_setting=1,
                model_ckpt=ckpt_pth,
            )
        elif '512' in ckpt_pth:
            editor = MMagicInferencer(
                'draggan',
                model_setting=0,
                model_ckpt=ckpt_pth,
            )
        elif '1024' in ckpt_pth:
            editor = MMagicInferencer(
                'draggan',
                model_setting=2,
                model_ckpt=ckpt_pth,
            )
        else:
            raise NotImplementedError
        self.editor = editor
        # Generate random latents.
        self.w0_seed = w0_seed
        self.w_load = w_load

        if self.w_load is None:
            # Generate random latents.
            z = torch.from_numpy(np.random.RandomState(w0_seed).randn(
                1, 512)).to(torch.float32).requires_grad_(True)

            # Run mapping network.
            sample_kwargs = {
                'truncation': trunc_psi,
                'return_noise': True,
                'return_features': True
            }
            extra_parameters = {
                'sample_kwargs': sample_kwargs,
                'num_batches': 1,
                'noise': z,
                'sample_model': 'ema',
                'infer_with_grad': True
            }
            results = self.editor.infer(extra_parameters=extra_parameters)

            w = results[0]['latent'].unsqueeze(0)  # [1, 16, n_dim]
        else:
            w = self.w_load.clone().to(self._device)

        self.w0 = w.detach().clone()
        self.w_plus = w_plus
        if w_plus:
            self.w = w.detach()
        else:
            self.w = w[:, 0, :].detach()
        self.w.requires_grad = True
        self.w_optim = torch.optim.Adam([self.w], lr=lr)

        self.feat_refs = None
        self.points0_pt = None

    def update_lr(self, lr):

        del self.w_optim
        self.w_optim = torch.optim.Adam([self.w], lr=lr)
        print(f'Rebuild optimizer with lr: {lr}')
        print('    Remain feat_refs and points0_pt')

    def _render_drag_impl(self,
                          res,
                          points=[],
                          targets=[],
                          mask=None,
                          lambda_mask=10,
                          reg=0,
                          feature_idx=5,
                          r1=3,
                          r2=12,
                          random_seed=0,
                          noise_mode='const',
                          trunc_psi=0.7,
                          force_fp32=False,
                          layer_name=None,
                          sel_channels=3,
                          base_channel=0,
                          img_scale_db=0,
                          img_normalize=False,
                          untransform=False,
                          is_drag=False,
                          reset=False,
                          to_pil=False,
                          **kwargs):
        ws = self.w
        if ws.dim() == 2:
            ws = ws.unsqueeze(1).repeat(1, 6, 1)
        ws = torch.cat([ws[:, :6, :], self.w0[:, 6:, :]],
                       dim=1)  # [1, 16, n_dim]
        if hasattr(self, 'points'):
            if len(points) != len(self.points):
                reset = True
        if reset:
            self.feat_refs = None
            self.points0_pt = None
        self.points = points

        # Run synthesis network.
        sample_kwargs = {
            'truncation': 1,
            'return_noise': True,
            'return_features': True,
            'input_is_latent': True
        }
        extra_parameters = {
            'sample_kwargs': sample_kwargs,
            'num_batches': 1,
            'noise': ws[0][:1],
            'sample_model': 'ema',
            'infer_with_grad': True
        }

        results = self.editor.infer(extra_parameters=extra_parameters)
        img = results[0]['fake_img'].unsqueeze(
            0) / 127.5 - 1.  # requires_grad==True
        feat_5 = results[0]['feats'].unsqueeze(0)  # requires_grad==True
        h, w = img.shape[-1], img.shape[-2]

        if is_drag:
            X = torch.linspace(0, h, h)
            Y = torch.linspace(0, w, w)
            xx, yy = torch.meshgrid(X, Y)
            feat_resize = F.interpolate(feat_5, [h, w], mode='bilinear')
            if self.feat_refs is None:
                self.feat0_resize = F.interpolate(
                    feat_5.detach(), [h, w], mode='bilinear')
                self.feat_refs = []
                for point in points:
                    py, px = round(point[0]), round(point[1])
                    self.feat_refs.append(self.feat0_resize[:, :, py, px])
                self.points0_pt = torch.Tensor(points).unsqueeze(0).to(
                    self._device)  # 1, N, 2

            # Point tracking with feature matching
            with torch.no_grad():
                for j, point in enumerate(points):
                    r = round(r2 / 512 * h)
                    up = max(point[0] - r, 0)
                    down = min(point[0] + r + 1, h)
                    left = max(point[1] - r, 0)
                    right = min(point[1] + r + 1, w)
                    feat_patch = feat_resize[:, :, up:down, left:right]
                    L2 = torch.linalg.norm(
                        feat_patch - self.feat_refs[j].reshape(1, -1, 1, 1),
                        dim=1)
                    _, idx = torch.min(L2.view(1, -1), -1)
                    width = right - left
                    point = [
                        idx.item() // width + up,
                        idx.item() % width + left
                    ]
                    points[j] = point

            res.points = [[point[0], point[1]] for point in points]

            # Motion supervision
            loss_motion = 0
            res.stop = True
            for j, point in enumerate(points):
                direction = torch.Tensor(
                    [targets[j][1] - point[1], targets[j][0] - point[0]])
                if torch.linalg.norm(direction) > max(2 / 512 * h, 2):
                    res.stop = False
                if torch.linalg.norm(direction) > 1:
                    distance = ((xx.to(self._device) - point[0])**2 +
                                (yy.to(self._device) - point[1])**2)**0.5
                    relis, reljs = torch.where(distance < round(r1 / 512 * h))
                    direction = direction / (
                        torch.linalg.norm(direction) + 1e-7)
                    gridh = (relis - direction[1]) / (h - 1) * 2 - 1
                    gridw = (reljs - direction[0]) / (w - 1) * 2 - 1
                    grid = torch.stack([gridw, gridh],
                                       dim=-1).unsqueeze(0).unsqueeze(0)
                    target = F.grid_sample(
                        feat_resize.float(), grid,
                        align_corners=True).squeeze(2)
                    loss_motion += F.l1_loss(feat_resize[:, :, relis, reljs],
                                             target.detach())

            loss = loss_motion
            if mask is not None:
                if mask.min() == 0 and mask.max() == 1:
                    mask_usq = mask.to(self._device).unsqueeze(0).unsqueeze(0)
                    loss_fix = F.l1_loss(feat_resize * mask_usq,
                                         self.feat0_resize * mask_usq)
                    loss += lambda_mask * loss_fix

            loss += reg * F.l1_loss(ws, self.w0)  # latent code regularization
            if not res.stop:
                self.w_optim.zero_grad()
                print(loss)
                loss.backward()
                self.w_optim.step()

        # Scale and convert to uint8.
        img = img[0]
        if img_normalize:
            img = img / img.norm(
                float('inf'), dim=[1, 2], keepdim=True).clip(1e-8, 1e8)
        img = img * (10**(img_scale_db / 20))
        img = (img * 127.5 + 128).clamp(0,
                                        255).to(torch.uint8).permute(1, 2, 0)
        if to_pil:
            from PIL import Image
            img = img.cpu().numpy()
            img = Image.fromarray(img)
        res.image = img
        res.w = ws.detach().cpu().numpy()



# Copyright (c) OpenMMLab. All rights reserved.
from .modified_vgg import ModifiedVGG
from .sr_resnet import MSRResNet
from .srgan import SRGAN

__all__ = [
    'ModifiedVGG',
    'MSRResNet',
    'SRGAN',
]



# Copyright (c) OpenMMLab. All rights reserved.
import os.path as osp
import warnings
from typing import Dict, List, Optional, Union

import torch
import yaml
from mmengine.registry import init_default_scope

from .inferencers import Inferencers
from .inferencers.base_mmagic_inferencer import InputsType


class MMagicInferencer:
    """MMagicInferencer API for mmagic models inference.

    Args:
        model_name (str): Name of the editing model.
        model_setting (str): Setting of a specific model.
            Default to 'a'.
        model_config (str): Path to the config file for the editing model.
            Default to None.
        model_ckpt (str): Path to the checkpoint file for the editing model.
            Default to None.
        config_dir (str): Path to the directory containing config files.
            Default to 'configs/'.
        device (torch.device): Device to use for inference. Default to 'cuda'.

    Examples:
        >>> # inference of a conditional model, biggan for example
        >>> editor = MMagicInferencer(model_name='biggan')
        >>> editor.infer(label=1, result_out_dir='./biggan_res.jpg')

        >>> # inference of a translation model, pix2pix for example
        >>> editor = MMagicInferencer(model_name='pix2pix')
        >>> editor.infer(img='./test.jpg', result_out_dir='./pix2pix_res.jpg')

        >>> # see demo/mmediting_inference_tutorial.ipynb for more examples
    """
    # unsupported now
    # singan, liif
    # output should be checked
    # dic, glean

    inference_supported_models = [
        # colorization models
        'inst_colorization',

        # conditional models
        'biggan',
        'sngan_proj',
        'sagan',

        # unconditional models
        'dcgan',
        'deblurganv2',
        'wgan-gp',
        'lsgan',
        'ggan',
        'pggan',
        'styleganv1',
        'styleganv2',
        'styleganv3',

        # matting models
        'dim',
        'indexnet',
        'gca',

        # inpainting models
        'aot_gan',
        'deepfillv1',
        'deepfillv2',
        'global_local',
        'partial_conv',

        # translation models
        'pix2pix',
        'cyclegan',

        # image super-resolution models
        'srcnn',
        'srgan_resnet',
        'edsr',
        'esrgan',
        'rdn',
        'dic',
        'ttsr',
        'glean',
        'real_esrgan',

        # video_interpolation models
        'flavr',
        'cain',

        # video_restoration models
        'edvr',
        'tdan',
        'basicvsr',
        'iconvsr',
        'basicvsr_pp',
        'real_basicvsr',

        # image_restoration models
        'nafnet',
        'swinir',
        'restormer',

        # text2image models
        'controlnet',
        'disco_diffusion',
        'stable_diffusion',

        # 3D-aware generation
        'eg3d',

        # animation inferencer
        'controlnet_animation',

        # draggan
        'draggan',

        # diffusers pipeline inferencer
        'diffusers_pipeline',
    ]

    inference_supported_models_cfg = {}
    inference_supported_models_cfg_inited = False

    def __init__(self,
                 model_name: str = None,
                 model_setting: int = None,
                 config_name: int = None,
                 model_config: str = None,
                 model_ckpt: str = None,
                 device: torch.device = None,
                 extra_parameters: Dict = None,
                 seed: int = 2022,
                 **kwargs) -> None:
        init_default_scope('mmagic')
        MMagicInferencer.init_inference_supported_models_cfg()
        inferencer_kwargs = {}
        inferencer_kwargs.update(
            self._get_inferencer_kwargs(model_name, model_setting, config_name,
                                        model_config, model_ckpt,
                                        extra_parameters))
        self.inferencer = Inferencers(
            device=device, seed=seed, **inferencer_kwargs)

    def _get_inferencer_kwargs(self, model_name: Optional[str],
                               model_setting: Optional[int],
                               config_name: Optional[int],
                               model_config: Optional[str],
                               model_ckpt: Optional[str],
                               extra_parameters: Optional[Dict]) -> Dict:
        """Get the kwargs for the inferencer."""
        kwargs = {}

        if model_name is not None:
            cfgs = self.get_model_config(model_name)
            kwargs['task'] = cfgs['task']
            setting_to_use = 0
            if model_setting:
                setting_to_use = model_setting
            config_dir = cfgs['settings'][setting_to_use]['Config']
            if config_name:
                for setting in cfgs['settings']:
                    if setting['Name'] == config_name:
                        config_dir = setting['Config']
                        break
            config_dir = config_dir[config_dir.find('configs'):]
            if osp.exists(
                    osp.join(osp.dirname(__file__), '..', '..', config_dir)):
                kwargs['config'] = osp.join(
                    osp.dirname(__file__), '..', '..', config_dir)
            else:
                kwargs['config'] = osp.join(
                    osp.dirname(__file__), '..', '.mim', config_dir)
            if 'Weights' in cfgs['settings'][setting_to_use].keys():
                kwargs['ckpt'] = cfgs['settings'][setting_to_use]['Weights']
                if model_name == 'controlnet':
                    kwargs['ckpt'] = None

        if model_config is not None:
            if kwargs.get('config', None) is not None:
                warnings.warn(
                    f'{model_name}\'s default config '
                    f'is overridden by {model_config}', UserWarning)
            kwargs['config'] = model_config

        if model_ckpt is not None:
            if kwargs.get('ckpt', None) is not None:
                warnings.warn(
                    f'{model_name}\'s default checkpoint '
                    f'is overridden by {model_ckpt}', UserWarning)
            kwargs['ckpt'] = model_ckpt

        if extra_parameters is not None:
            kwargs['extra_parameters'] = extra_parameters

        return kwargs

    def print_extra_parameters(self):
        """Print the unique parameters of each kind of inferencer."""
        extra_parameters = self.inferencer.get_extra_parameters()
        print(extra_parameters)

    def infer(self,
              img: InputsType = None,
              video: InputsType = None,
              label: InputsType = None,
              trimap: InputsType = None,
              mask: InputsType = None,
              result_out_dir: str = '',
              **kwargs) -> Union[Dict, List[Dict]]:
        """Infer edit model on an image(video).

        Args:
            img (str): Img path.
            video (str): Video path.
            label (int): Label for conditional or unconditional models.
            trimap (str): Trimap path for matting models.
            mask (str): Mask path for inpainting models.
            result_out_dir (str): Output directory of result image or video.
                Defaults to ''.

        Returns:
            Dict or List[Dict]: Each dict contains the inference result of
            each image or video.
        """
        return self.inferencer(
            img=img,
            video=video,
            label=label,
            trimap=trimap,
            mask=mask,
            result_out_dir=result_out_dir,
            **kwargs)

    def get_model_config(self, model_name: str) -> Dict:
        """Get the model configuration including model config and checkpoint
        url.

        Args:
            model_name (str): Name of the model.
        Returns:
            dict: Model configuration.
        """
        if model_name not in self.inference_supported_models:
            raise ValueError(f'Model {model_name} is not supported.')
        else:
            return self.inference_supported_models_cfg[model_name]

    @staticmethod
    def init_inference_supported_models_cfg() -> None:
        if not MMagicInferencer.inference_supported_models_cfg_inited:
            if osp.exists(
                    osp.join(osp.dirname(__file__), '..', '..', 'configs')):
                all_cfgs_dir = osp.join(
                    osp.dirname(__file__), '..', '..', 'configs')
            else:
                all_cfgs_dir = osp.join(
                    osp.dirname(__file__), '..', '.mim', 'configs')
            for model_name in MMagicInferencer.inference_supported_models:
                meta_file_dir = osp.join(all_cfgs_dir, model_name,
                                         'metafile.yml')
                with open(meta_file_dir, 'r') as stream:
                    parsed_yaml = yaml.safe_load(stream)
                task = parsed_yaml['Models'][0]['Results'][0]['Task']
                MMagicInferencer.inference_supported_models_cfg[
                    model_name] = {}
                MMagicInferencer.inference_supported_models_cfg[model_name][
                    'task'] = task  # noqa
                MMagicInferencer.inference_supported_models_cfg[model_name][
                    'settings'] = parsed_yaml['Models']  # noqa

            MMagicInferencer.inference_supported_models_cfg_inited = True

    @staticmethod
    def get_inference_supported_models() -> List:
        """static function for getting inference supported modes."""
        return MMagicInferencer.inference_supported_models

    @staticmethod
    def get_inference_supported_tasks() -> List:
        """static function for getting inference supported tasks."""
        if not MMagicInferencer.inference_supported_models_cfg_inited:
            MMagicInferencer.init_inference_supported_models_cfg()

        supported_task = set()
        for key in MMagicInferencer.inference_supported_models_cfg.keys():
            if MMagicInferencer.inference_supported_models_cfg[key]['task'] \
               not in supported_task:
                supported_task.add(MMagicInferencer.
                                   inference_supported_models_cfg[key]['task'])
        return list(supported_task)

    @staticmethod
    def get_task_supported_models(task: str) -> List:
        """static function for getting task supported models."""
        if not MMagicInferencer.inference_supported_models_cfg_inited:
            MMagicInferencer.init_inference_supported_models_cfg()

        supported_models = []
        for key in MMagicInferencer.inference_supported_models_cfg.keys():
            if MMagicInferencer.inference_supported_models_cfg[key][
                    'task'] == task:
                supported_models.append(key)
        return supported_models



_base_ = [
    '../_base_/datasets/unconditional_imgs_64x64.py',
    '../_base_/gen_default_runtime.py'
]

model = dict(
    type='GGAN',
    noise_size=1024,
    data_preprocessor=dict(type='DataPreprocessor'),
    generator=dict(type='LSGANGenerator', output_scale=64),
    discriminator=dict(type='LSGANDiscriminator', input_scale=64))

# define dataset
batch_size = 128
data_root = 'data/lsun/images/bedroom_train'
train_dataloader = dict(
    batch_size=batch_size, dataset=dict(data_root=data_root))

val_dataloader = dict(batch_size=batch_size, dataset=dict(data_root=data_root))

test_dataloader = dict(
    batch_size=batch_size, dataset=dict(data_root=data_root))

optim_wrapper = dict(
    generator=dict(optimizer=dict(type='Adam', lr=0.0001, betas=(0.5, 0.99))),
    discriminator=dict(
        optimizer=dict(type='Adam', lr=0.0001, betas=(0.5, 0.99))))

default_hooks = dict(
    checkpoint=dict(
        max_keep_ckpts=20,
        save_best=['FID-Full-50k/fid', 'swd/avg', 'ms-ssim/avg'],
        rule=['less', 'less', 'greater']))

# VIS_HOOK
custom_hooks = [
    dict(
        type='VisualizationHook',
        interval=5000,
        fixed_input=True,
        vis_kwargs_list=dict(type='GAN', name='fake_img'))
]

train_cfg = dict(max_iters=160000)

# METRICS
metrics = [
    dict(
        type='FrechetInceptionDistance',
        prefix='FID-Full-50k',
        fake_nums=50000,
        inception_style='StyleGAN',
        sample_model='orig'),
    dict(
        type='MS_SSIM', prefix='ms-ssim', fake_nums=10000,
        sample_model='orig'),
    dict(
        type='SWD',
        prefix='swd',
        fake_nums=16384,
        sample_model='orig',
        image_shape=(3, 64, 64))
]

val_metrics = [
    dict(
        type='FrechetInceptionDistance',
        prefix='FID-Full-50k',
        fake_nums=50000,
        inception_style='StyleGAN',
        sample_model='orig'),
]

val_evaluator = dict(metrics=val_metrics)
test_evaluator = dict(metrics=metrics)



_base_ = './glean_in128out1024_4xb2-300k_ffhq-celeba-hq.py'

experiment_name = 'glean_in128out1024-fp16_4xb2-300k_ffhq-celeba-hq'
work_dir = f'./work_dirs/{experiment_name}'
save_dir = './work_dirs'

# model settings
model = dict(
    generator=dict(fp16_enabled=True), discriminator=dict(fp16_enabled=True))



# Copyright (c) OpenMMLab. All rights reserved.
from typing import List, Tuple

from mmcv.transforms.base import BaseTransform

from mmagic.registry import TRANSFORMS
from mmagic.structures import DataSample
from mmagic.utils import all_to_tensor


@TRANSFORMS.register_module()
class PackInputs(BaseTransform):
    """Pack data into DataSample for training, evaluation and testing.

    MMagic follows the design of data structure from MMEngine.
        Data from the loader will be packed into data field of DataSample.
        More details of DataSample refer to the documentation of MMEngine:
        https://mmengine.readthedocs.io/en/latest/advanced_tutorials/data_element.html

    Args:
        keys Tuple[List[str], str, None]: The keys to saved in returned
            inputs, which are used as the input of models, default to
            ['img', 'noise', 'merged'].
        data_keys Tuple[List[str], str, None]: The keys to saved in
            `data_field` of the `data_samples`.
        meta_keys Tuple[List[str], str, None]: The meta keys to saved
            in `metainfo` of the `data_samples`. All the other data will
            be packed into the data of the `data_samples`
    """

    def __init__(
        self,
        keys: Tuple[List[str], str] = ['merged', 'img'],
        meta_keys: Tuple[List[str], str] = [],
        data_keys: Tuple[List[str], str] = [],
    ) -> None:

        assert keys is not None, \
            'keys in PackInputs can not be None.'
        assert data_keys is not None, \
            'data_keys in PackInputs can not be None.'
        assert meta_keys is not None, \
            'meta_keys in PackInputs can not be None.'

        self.keys = keys if isinstance(keys, List) else [keys]
        self.data_keys = data_keys if isinstance(data_keys,
                                                 List) else [data_keys]
        self.meta_keys = meta_keys if isinstance(meta_keys,
                                                 List) else [meta_keys]

    def transform(self, results: dict) -> dict:
        """Method to pack the input data.

        Args:
            results (dict): Result dict from the data pipeline.

        Returns:
            dict: A dict contains

            - 'inputs' (obj:`dict`): The forward data of models.
              According to different tasks, the `inputs` may contain images,
              videos, labels, text, etc.

            - 'data_samples' (obj:`DataSample`): The annotation info of the
                sample.
        """

        # prepare inputs
        inputs = dict()
        for k in self.keys:
            value = results.get(k, None)
            if value is not None:
                inputs[k] = all_to_tensor(value)

        # return the inputs as tensor, if it has only one item
        if len(inputs.values()) == 1:
            inputs = list(inputs.values())[0]

        data_sample = DataSample()
        # prepare metainfo and data in DataSample according to predefined keys
        predefined_data = {
            k: v
            for (k, v) in results.items()
            if k not in (self.data_keys + self.meta_keys)
        }
        data_sample.set_predefined_data(predefined_data)

        # prepare metainfo in DataSample according to user-provided meta_keys
        required_metainfo = {
            k: v
            for (k, v) in results.items() if k in self.meta_keys
        }
        data_sample.set_metainfo(required_metainfo)

        # prepare metainfo in DataSample according to user-provided data_keys
        required_data = {
            k: v
            for (k, v) in results.items() if k in self.data_keys
        }
        data_sample.set_tensor_data(required_data)

        return {'inputs': inputs, 'data_samples': data_sample}

    def __repr__(self) -> str:

        repr_str = self.__class__.__name__

        return repr_str



# Copyright (c) OpenMMLab. All rights reserved.
from copy import deepcopy
from functools import partial

import mmengine
import torch
import torch.nn as nn
from mmcv.ops.fused_bias_leakyrelu import fused_bias_leakyrelu
from mmcv.ops.upfirdn2d import upfirdn2d
from mmengine.model import BaseModule

from mmagic.registry import MODELS
from ..pggan import (EqualizedLRConvModule, EqualizedLRConvUpModule,
                     EqualizedLRLinearModule)


class EqualLinearActModule(BaseModule):
    """Equalized LR Linear Module with Activation Layer.

    This module is modified from ``EqualizedLRLinearModule`` defined in PGGAN.
    The major features updated in this module is adding support for activation
    layers used in StyleGAN2.

    Args:
        equalized_lr_cfg (dict | None, optional): Config for equalized lr.
            Defaults to dict(gain=1., lr_mul=1.).
        bias (bool, optional): Whether to use bias item. Defaults to True.
        bias_init (float, optional): The value for bias initialization.
            Defaults to ``0.``.
        act_cfg (dict | None, optional): Config for activation layer.
            Defaults to None.
    """

    def __init__(self,
                 *args,
                 equalized_lr_cfg=dict(gain=1., lr_mul=1.),
                 bias=True,
                 bias_init=0.,
                 act_cfg=None,
                 **kwargs):
        super().__init__()
        self.with_activation = act_cfg is not None
        # w/o bias in linear layer
        self.linear = EqualizedLRLinearModule(
            *args, bias=False, equalized_lr_cfg=equalized_lr_cfg, **kwargs)

        if equalized_lr_cfg is not None:
            self.lr_mul = equalized_lr_cfg.get('lr_mul', 1.)
        else:
            self.lr_mul = 1.

        # define bias outside linear layer
        if bias:
            self.bias = nn.Parameter(
                torch.zeros(self.linear.out_features).fill_(bias_init))
        else:
            self.bias = None

        if self.with_activation:
            act_cfg = deepcopy(act_cfg)
            if act_cfg['type'] == 'fused_bias':
                self.act_type = act_cfg.pop('type')
                assert self.bias is not None
                self.activate = partial(fused_bias_leakyrelu, **act_cfg)
            else:
                self.act_type = 'normal'
                self.activate = MODELS.build(act_cfg)
        else:
            self.act_type = None

    def forward(self, x):
        """Forward function.

        Args:
            x (Tensor): Input feature map with shape of (N, C, ...).

        Returns:
            Tensor: Output feature map.
        """
        if x.ndim >= 3:
            x = x.reshape(x.size(0), -1)
        x = self.linear(x)

        if self.with_activation and self.act_type == 'fused_bias':
            x = self.activate(x, self.bias * self.lr_mul)
        elif self.bias is not None and self.with_activation:
            x = self.activate(x + self.bias * self.lr_mul)
        elif self.bias is not None:
            x = x + self.bias * self.lr_mul
        elif self.with_activation:
            x = self.activate(x)

        return x


class NoiseInjection(BaseModule):
    """Noise Injection Module.

    In StyleGAN2, they adopt this module to inject spatial random noise map in
    the generators.

    Args:
        noise_weight_init (float, optional): Initialization weight for noise
            injection. Defaults to ``0.``.
        fixed_noise (bool, optional): Whether to inject a fixed noise. Defaults
        to ``False``.
    """

    def __init__(self, noise_weight_init=0., fixed_noise=False):
        super().__init__()
        self.weight = nn.Parameter(torch.zeros(1).fill_(noise_weight_init))
        self.fixed_noise = fixed_noise

    def forward(self, image, noise=None, return_noise=False):
        """Forward Function.

        Args:
            image (Tensor): Spatial features with a shape of (N, C, H, W).
            noise (Tensor, optional): Noises from the outside.
                Defaults to None.
            return_noise (bool, optional): Whether to return noise tensor.
                Defaults to False.

        Returns:
            Tensor: Output features.
        """
        if noise is None:
            batch, _, height, width = image.shape
            noise = image.new_empty(batch, 1, height, width).normal_()
            if self.fixed_noise:
                torch.manual_seed(1024)
                noise = torch.randn(batch, 1, height, width).cuda()

        noise = noise.to(image.dtype)
        if return_noise:
            return image + self.weight.to(image.dtype) * noise, noise

        return image + self.weight.to(image.dtype) * noise


class ConstantInput(BaseModule):
    """Constant Input.

    In StyleGAN2, they substitute the original head noise input with such a
    constant input module.

    Args:
        channel (int): Channels for the constant input tensor.
        size (int, optional): Spatial size for the constant input.
            Defaults to 4.
    """

    def __init__(self, channel, size=4):
        super().__init__()
        if isinstance(size, int):
            size = [size, size]
        elif mmengine.is_seq_of(size, int):
            assert len(
                size
            ) == 2, f'The length of size should be 2 but got {len(size)}'
        else:
            raise ValueError(f'Got invalid value in size, {size}')

        self.input = nn.Parameter(torch.randn(1, channel, *size))

    def forward(self, x):
        """Forward function.

        Args:
            x (Tensor): Input feature map with shape of (N, C, ...).

        Returns:
            Tensor: Output feature map.
        """
        batch = x.shape[0]
        out = self.input.repeat(batch, 1, 1, 1)

        return out


def make_kernel(k):
    k = torch.tensor(k, dtype=torch.float32)
    if k.ndim == 1:
        k = k[None, :] * k[:, None]

    k /= k.sum()

    return k


class Blur(BaseModule):
    """Blur module.

    This module is adopted rightly after upsampling operation in StyleGAN2.

    Args:
        kernel (Array): Blur kernel/filter used in UpFIRDn.
        pad (list[int]): Padding for features.
        upsample_factor (int, optional): Upsampling factor. Defaults to 1.
    """

    def __init__(self, kernel, pad, upsample_factor=1):
        super().__init__()
        kernel = make_kernel(kernel)
        if upsample_factor > 1:
            kernel = kernel * (upsample_factor**2)

        self.register_buffer('kernel', kernel)

        self.pad = pad

    def forward(self, x):
        """Forward function.

        Args:
            x (Tensor): Input feature map with shape of (N, C, H, W).

        Returns:
            Tensor: Output feature map.
        """

        # In Tero's implementation, he uses fp32
        return upfirdn2d(x, self.kernel.to(x.dtype), padding=self.pad)


class AdaptiveInstanceNorm(BaseModule):
    r"""Adaptive Instance Normalization Module.

    Ref: https://github.com/rosinality/style-based-gan-pytorch/blob/master/model.py  # noqa

    Args:
        in_channel (int): The number of input's channel.
        style_dim (int): Style latent dimension.
    """

    def __init__(self, in_channel, style_dim):
        super().__init__()

        self.norm = nn.InstanceNorm2d(in_channel)
        self.affine = EqualizedLRLinearModule(style_dim, in_channel * 2)

        self.affine.bias.data[:in_channel] = 1
        self.affine.bias.data[in_channel:] = 0

    def forward(self, input, style):
        """Forward function.

        Args:
            input (Tensor): Input tensor with shape (n, c, h, w).
            style (Tensor): Input style tensor with shape (n, c).

        Returns:
            Tensor: Forward results.
        """
        style = self.affine(style).unsqueeze(2).unsqueeze(3)
        gamma, beta = style.chunk(2, 1)

        out = self.norm(input)
        out = gamma * out + beta

        return out


class StyleConv(BaseModule):

    def __init__(self,
                 in_channels,
                 out_channels,
                 kernel_size,
                 style_channels,
                 padding=1,
                 initial=False,
                 blur_kernel=[1, 2, 1],
                 upsample=False,
                 fused=False):
        """Convolutional style blocks composing of noise injector, AdaIN module
        and convolution layers.

        Args:
            in_channels (int): The channel number of the input tensor.
            out_channels (itn): The channel number of the output tensor.
            kernel_size (int): The kernel size of convolution layers.
            style_channels (int): The number of channels for style code.
            padding (int, optional): Padding of convolution layers.
                Defaults to 1.
            initial (bool, optional): Whether this is the first StyleConv of
                StyleGAN's generator. Defaults to False.
            blur_kernel (list, optional): The blurry kernel.
                Defaults to [1, 2, 1].
            upsample (bool, optional): Whether perform upsampling.
                Defaults to False.
            fused (bool, optional): Whether use fused upconv.
                Defaults to False.
        """
        super().__init__()

        if initial:
            self.conv1 = ConstantInput(in_channels)
        else:
            if upsample:
                if fused:
                    self.conv1 = nn.Sequential(
                        EqualizedLRConvUpModule(
                            in_channels,
                            out_channels,
                            kernel_size,
                            padding=padding,
                            act_cfg=dict(type='LeakyReLU',
                                         negative_slope=0.2)),
                        Blur(blur_kernel, pad=(1, 1)),
                    )
                else:
                    self.conv1 = nn.Sequential(
                        nn.Upsample(scale_factor=2, mode='nearest'),
                        EqualizedLRConvModule(
                            in_channels,
                            out_channels,
                            kernel_size,
                            padding=padding,
                            act_cfg=None), Blur(blur_kernel, pad=(1, 1)))
            else:
                self.conv1 = EqualizedLRConvModule(
                    in_channels,
                    out_channels,
                    kernel_size,
                    padding=padding,
                    act_cfg=None)

        self.noise_injector1 = NoiseInjection()
        self.activate1 = nn.LeakyReLU(0.2)
        self.adain1 = AdaptiveInstanceNorm(out_channels, style_channels)

        self.conv2 = EqualizedLRConvModule(
            out_channels,
            out_channels,
            kernel_size,
            padding=padding,
            act_cfg=None)
        self.noise_injector2 = NoiseInjection()
        self.activate2 = nn.LeakyReLU(0.2)
        self.adain2 = AdaptiveInstanceNorm(out_channels, style_channels)

    def forward(self,
                x,
                style1,
                style2,
                noise1=None,
                noise2=None,
                return_noise=False):
        """Forward function.

        Args:
            x (Tensor): Input tensor.
            style1 (Tensor): Input style tensor with shape (n, c).
            style2 (Tensor): Input style tensor with shape (n, c).
            noise1 (Tensor, optional): Noise tensor with shape (n, c, h, w).
                Defaults to None.
            noise2 (Tensor, optional): Noise tensor with shape (n, c, h, w).
                Defaults to None.
            return_noise (bool, optional): If True, ``noise1`` and ``noise2``
            will be returned with ``out``. Defaults to False.

        Returns:
            Tensor | tuple[Tensor]: Forward results.
        """
        out = self.conv1(x)
        if return_noise:
            out, noise1 = self.noise_injector1(
                out, noise=noise1, return_noise=return_noise)
        else:
            out = self.noise_injector1(
                out, noise=noise1, return_noise=return_noise)
        out = self.activate1(out)
        out = self.adain1(out, style1)

        out = self.conv2(out)
        if return_noise:
            out, noise2 = self.noise_injector2(
                out, noise=noise2, return_noise=return_noise)
        else:
            out = self.noise_injector2(
                out, noise=noise2, return_noise=return_noise)
        out = self.activate2(out)
        out = self.adain2(out, style2)

        if return_noise:
            return out, noise1, noise2

        return out



# Copyright (c) OpenMMLab. All rights reserved.
import copy
import inspect
from typing import List

import numpy as np
from mmcv.transforms import BaseTransform

from mmagic.registry import TRANSFORMS

try:
    import albumentations
    from albumentations import Compose
except ImportError:
    albumentations = None
    Compose = None


@TRANSFORMS.register_module()
class Albumentations(BaseTransform):
    """Albumentation augmentation.

    Adds custom transformations from Albumentations library.
    Please, visit `https://github.com/albumentations-team/albumentations`
    and `https://albumentations.ai/docs/getting_started/transforms_and_targets`
    to get more information.

    An example of ``transforms`` is as followed:

    .. code-block::

        albu_transforms = [
            dict(
                type='Resize',
                height=100,
                width=100,
            ),
            dict(
                type='RandomFog',
                p=0.5,
            ),
            dict(
                type='RandomRain',
                p=0.5
            ),
            dict(
                type='RandomSnow',
                p=0.5,
            ),
        ]
        pipeline = [
            dict(
                type='LoadImageFromFile',
                key='img',
                color_type='color',
                channel_order='rgb',
                imdecode_backend='cv2'),
            dict(
                type='Albumentations',
                keys=['img'],
                transforms=albu_transforms),
            dict(type='PackInputs')
        ]

    Args:
        keys (list[str]): A list specifying the keys whose values are modified.
        transforms (list[dict]): A list of albu transformations.
    """

    def __init__(self, keys: List[str], transforms: List[dict]) -> None:

        if Compose is None:
            raise RuntimeError('Please install albumentations')

        self.keys = keys

        # Args will be modified later, copying it will be safer
        transforms = copy.deepcopy(transforms)
        self.transforms = transforms
        self.aug = Compose([self.albu_builder(t) for t in self.transforms])

    def albu_builder(self, cfg: dict) -> albumentations:
        """Import a module from albumentations.

        It inherits some of :func:`build_from_cfg` logic.

        Args:
            cfg (dict): Config dict. It should at least contain the key "type".

        Returns:
            obj: The constructed object.
        """

        assert isinstance(cfg, dict) and 'type' in cfg
        args = cfg.copy()
        obj_type = args.pop('type')
        if isinstance(obj_type, str):
            if albumentations is None:
                raise RuntimeError('Please install albumentations')
            obj_cls = getattr(albumentations, obj_type)
        elif inspect.isclass(obj_type):
            obj_cls = obj_type
        else:
            raise TypeError(
                f'type must be a str or valid type, but got {type(obj_type)}')

        if 'transforms' in args:
            args['transforms'] = [
                self.albu_builder(transform)
                for transform in args['transforms']
            ]

        return obj_cls(**args)

    def _apply_albu(self, imgs):
        is_single_image = False
        if isinstance(imgs, np.ndarray):
            is_single_image = True
            imgs = [imgs]

        outputs = []
        for img in imgs:
            outputs.append(self.aug(image=img)['image'])

        if is_single_image:
            outputs = outputs[0]

        return outputs

    def transform(self, results):
        """Transform function of Albumentations."""

        for k in self.keys:
            results[k] = self._apply_albu(results[k])

        return results

    def __repr__(self):
        repr_str = self.__class__.__name__
        repr_str += f'(keys={self.keys}, transforms={self.transforms})'

        return repr_str
